
# KlippyConnection - manage unix socket connection to Klipper
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import os
import time
import logging
import getpass
import asyncio
import pathlib
from ..utils import ServerError, get_unix_peer_credentials
from ..utils import json_wrapper as jsonw
from ..common import KlippyState, RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Awaitable,
    Optional,
    Callable,
    Coroutine,
    Dict,
    List,
    Set,
    Tuple,
    Union
)
if TYPE_CHECKING:
    from ..common import WebRequest, APITransport, BaseRemoteConnection
    from ..confighelper import ConfigHelper
    from .klippy_apis import KlippyAPI
    from .file_manager.file_manager import FileManager
    from .machine import Machine
    from .job_state import JobState
    from .database import MoonrakerDatabase as Database
    FlexCallback = Callable[..., Optional[Coroutine]]
    Subscription = Dict[str, Optional[List[str]]]

# These endpoints are reserved for klippy/moonraker communication only and are
# not exposed via http or the websocket
RESERVED_ENDPOINTS = [
    "list_endpoints",
    "gcode/subscribe_output",
    "register_remote_method",
]

# Items to exclude from the subscription cache.  They never change and can be
# quite large.
CACHE_EXCLUSIONS = {
    "configfile": ["config", "settings"]
}

INIT_TIME = .25
LOG_ATTEMPT_INTERVAL = int(2. / INIT_TIME + .5)
MAX_LOG_ATTEMPTS = 10 * LOG_ATTEMPT_INTERVAL
UNIX_BUFFER_LIMIT = 20 * 1024 * 1024
SVC_INFO_KEY = "klippy_connection.service_info"
SRC_PATH_KEY = "klippy_connection.path"
PY_EXEC_KEY = "klippy_connection.executable"

class KlippyConnection:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.uds_address = config.getpath(
            "klippy_uds_address", pathlib.Path("/tmp/klippy_uds")
        )
        self.writer: Optional[asyncio.StreamWriter] = None
        self.connection_mutex: asyncio.Lock = asyncio.Lock()
        self.event_loop = self.server.get_event_loop()
        self.log_no_access = True
        # Connection State
        self.connection_task: Optional[asyncio.Task] = None
        self.closing: bool = False
        self.subscription_lock = asyncio.Lock()
        self._klippy_info: Dict[str, Any] = {}
        self._klippy_identified: bool = False
        self._klippy_initializing: bool = False
        self._klippy_started: bool = False
        self._methods_registered: bool = False
        self._klipper_version: str = ""
        self._missing_reqs: Set[str] = set()
        self._peer_cred: Dict[str, int] = {}
        self._service_info: Dict[str, Any] = {}
        self._path = pathlib.Path("~/klipper").expanduser()
        self._executable = pathlib.Path("~/klippy-env/bin/python").expanduser()
        self.init_attempts: int = 0
        self._state: KlippyState = KlippyState.DISCONNECTED
        self._state.set_message("Klippy Disconnected")
        self.subscriptions: Dict[APITransport, Subscription] = {}
        self.subscription_cache: Dict[str, Dict[str, Any]] = {}
        # Setup remote methods accessable to Klippy.  Note that all
        # registered remote methods should be of the notification type,
        # they do not return a response to Klippy after execution
        self.pending_requests: Dict[int, KlippyRequest] = {}
        self.remote_methods: Dict[str, FlexCallback] = {}
        self.klippy_reg_methods: List[str] = []
        self.register_remote_method(
            'process_gcode_response', self._process_gcode_response,
            need_klippy_reg=False)
        self.register_remote_method(
            'process_status_update', self._process_status_update,
            need_klippy_reg=False)

    @property
    def klippy_apis(self) -> KlippyAPI:
        return self.server.lookup_component("klippy_apis")

    @property
    def state(self) -> KlippyState:
        if self.is_connected() and not self._klippy_started:
            return KlippyState.STARTUP
        return self._state

    @property
    def state_message(self) -> str:
        return self._state.message

    @property
    def klippy_info(self) -> Dict[str, Any]:
        return self._klippy_info

    @property
    def missing_requirements(self) -> List[str]:
        return list(self._missing_reqs)

    @property
    def peer_credentials(self) -> Dict[str, int]:
        return dict(self._peer_cred)

    @property
    def service_info(self) -> Dict[str, Any]:
        return self._service_info

    @property
    def unit_name(self) -> str:
        svc_info = self._service_info
        unit_name = svc_info.get("unit_name", "klipper.service")
        return unit_name.split(".", 1)[0]

    @property
    def path(self) -> pathlib.Path:
        return self._path

    @property
    def executable(self) -> pathlib.Path:
        return self._executable

    def load_saved_state(self) -> None:
        db: Database = self.server.lookup_component("database")
        sync_provider = db.get_provider_wrapper()
        kc_info: Dict[str, Any]
        kc_info = sync_provider.get_item("moonraker", "klippy_connection", {})
        self._path = pathlib.Path(kc_info.get("path", str(self._path)))
        self._executable = pathlib.Path(kc_info.get("executable", str(self.executable)))
        self._service_info = kc_info.get("service_info", {})

    async def component_init(self) -> None:
        machine: Machine = self.server.lookup_component("machine")
        if self._service_info:
            machine.log_service_info(self._service_info)

    async def wait_connected(self) -> bool:
        if (
            self.connection_task is None or
            self.connection_task.done()
        ):
            return self.is_connected()
        try:
            await self.connection_task
        except Exception:
            pass
        return self.is_connected()

    async def wait_started(self, timeout: float = 20.) -> bool:
        if self.connection_task is None or not self.is_connected():
            return False
        if not self.connection_task.done():
            await asyncio.wait_for(
                asyncio.shield(self.connection_task), timeout=timeout)
        return self.is_connected()

    async def _read_stream(self, reader: asyncio.StreamReader) -> None:
        errors_remaining: int = 10
        while not reader.at_eof():
            try:
                data = await reader.readuntil(b'\x03')
            except (ConnectionError, asyncio.IncompleteReadError):
                break
            except asyncio.CancelledError:
                logging.exception("Klippy Stream Read Cancelled")
                raise
            except Exception:
                logging.exception("Klippy Stream Read Error")
                errors_remaining -= 1
                if not errors_remaining or not self.is_connected():
                    break
                continue
            errors_remaining = 10
            try:
                decoded_cmd = jsonw.loads(data[:-1])
                self._process_command(decoded_cmd)
            except Exception:
                logging.exception(
                    f"Error processing Klippy Host Response: {data.decode()}")
        if not self.closing:
            logging.debug("Klippy Disconnection From _read_stream()")
            await self.close()

    async def _write_request(self, request: KlippyRequest) -> None:
        if self.writer is None or self.closing:
            request.set_exception(ServerError("Klippy Host not connected", 503))
            return
        data = jsonw.dumps(request.to_dict()) + b"\x03"
        try:
            self.writer.write(data)
            await self.writer.drain()
        except asyncio.CancelledError:
            request.set_exception(ServerError("Klippy Write Request Cancelled", 503))
            raise
        except Exception:
            request.set_exception(ServerError("Klippy Write Request Error", 503))
            if not self.closing:
                logging.debug("Klippy Disconnection From _write_request()")
                await self.close()

    def register_remote_method(self,
                               method_name: str,
                               cb: FlexCallback,
                               need_klippy_reg: bool = True
                               ) -> None:
        if method_name in self.remote_methods:
            raise self.server.error(
                f"Remote method ({method_name}) already registered")
        if self.server.is_running():
            raise self.server.error(
                f"Failed to register remote method {method_name}, "
                "methods must be registered during initialization")
        self.remote_methods[method_name] = cb
        if need_klippy_reg:
            # These methods need to be registered with Klippy
            self.klippy_reg_methods.append(method_name)

    def register_method_from_agent(
        self, connection: BaseRemoteConnection, method_name: str
    ) -> Optional[Awaitable]:
        if connection.client_data["type"] != "agent":
            raise self.server.error(
                "Only connections of the 'agent' type can register methods"
            )
        if method_name in self.remote_methods:
            raise self.server.error(
                f"Remote method ({method_name}) already registered"
            )

        def _on_agent_method_received(**kwargs) -> None:
            connection.call_method(method_name, kwargs)
        self.remote_methods[method_name] = _on_agent_method_received
        self.klippy_reg_methods.append(method_name)
        if self._methods_registered and self._state != KlippyState.DISCONNECTED:
            coro = self.klippy_apis.register_method(method_name)
            return self.event_loop.create_task(coro)
        return None

    def unregister_method(self, method_name: str):
        self.remote_methods.pop(method_name, None)
        try:
            self.klippy_reg_methods.remove(method_name)
        except ValueError:
            pass

    def connect(self) -> Awaitable[bool]:
        if (
            self.is_connected() or
            not self.server.is_running() or
            (self.connection_task is not None and
             not self.connection_task.done())
        ):
            # already connecting
            fut = self.event_loop.create_future()
            fut.set_result(self.is_connected())
            return fut
        self.connection_task = self.event_loop.create_task(self._do_connect())
        return self.connection_task  # type: ignore

    async def _do_connect(self) -> bool:
        async with self.connection_mutex:
            while self.writer is None:
                await asyncio.sleep(INIT_TIME)
                if self.closing or not self.server.is_running():
                    return False
                if not self.uds_address.exists():
                    continue
                if not os.access(str(self.uds_address), os.R_OK | os.W_OK):
                    if self.log_no_access:
                        user = getpass.getuser()
                        logging.info(
                            f"Cannot connect to Klippy, Linux user '{user}' "
                            "lacks permission to open Unix Domain Socket: "
                            f"{self.uds_address}")
                        self.log_no_access = False
                    continue
                self.log_no_access = True
                try:
                    reader, writer = await self.open_klippy_connection(True)
                except asyncio.CancelledError:
                    raise
                except Exception:
                    continue
                logging.info("Klippy Connection Established")
                self.writer = writer
                if self._get_peer_credentials(writer):
                    await self._get_service_info(self._peer_cred["process_id"])
                self.event_loop.create_task(self._read_stream(reader))
            return await self._init_klippy_connection()

    async def open_klippy_connection(
        self, primary: bool = False
    ) -> Tuple[asyncio.StreamReader, asyncio.StreamWriter]:
        if not primary and not self.is_connected():
            raise ServerError("Klippy Unix Connection Not Available", 503)
        return await asyncio.open_unix_connection(
            str(self.uds_address), limit=UNIX_BUFFER_LIMIT)

    def _get_peer_credentials(self, writer: asyncio.StreamWriter) -> bool:
        peer_cred = get_unix_peer_credentials(writer, "Klippy")
        if not peer_cred:
            return False
        if peer_cred.get("process_id") == 1:
            logging.debug("Klipper Unix Socket created via Systemd Socket Activation")
            return False
        self._peer_cred = peer_cred
        logging.debug(
            f"Klippy Connection: Received Peer Credentials: {self._peer_cred}"
        )
        return True

    async def _get_service_info(self, process_id: int) -> None:
        machine: Machine = self.server.lookup_component("machine")
        provider = machine.get_system_provider()
        svc_info = await provider.extract_service_info("klipper", process_id)
        if svc_info != self._service_info:
            db: Database = self.server.lookup_component('database')
            db.insert_item("moonraker", SVC_INFO_KEY, svc_info)
            self._service_info = svc_info
            machine.log_service_info(svc_info)

    def _save_path_info(self) -> None:
        kpath = pathlib.Path(self._klippy_info["klipper_path"])
        kexec = pathlib.Path(self._klippy_info["python_path"])
        db: Database = self.server.lookup_component("database")
        if kpath != self.path:
            self._path = kpath
            db.insert_item("moonraker", SRC_PATH_KEY, str(kpath))
            logging.info(f"Updated Stored Klipper Source Path: {kpath}")
        if kexec != self.executable:
            self._executable = kexec
            db.insert_item("moonraker", PY_EXEC_KEY, str(kexec))
            logging.info(f"Updated Stored Klipper Python Path: {kexec}")

    async def _init_klippy_connection(self) -> bool:
        self._klippy_identified = False
        self._klippy_started = False
        self._klippy_initializing = True
        self._methods_registered = False
        self._missing_reqs.clear()
        self.init_attempts = 0
        self._state = KlippyState.STARTUP
        while self.server.is_running():
            await asyncio.sleep(INIT_TIME)
            await self._check_ready()
            if not self._klippy_initializing:
                logging.debug("Klippy Connection Initialized")
                return True
            if not self.is_connected():
                self._klippy_initializing = False
                break
            else:
                self.init_attempts += 1
        logging.debug("Klippy Connection Failed to Init")
        return False

    async def _request_endpoints(self) -> None:
        result = await self.klippy_apis.list_endpoints(default=None)
        if result is None:
            return
        endpoints = result.get('endpoints', [])
        for ep in endpoints:
            if ep not in RESERVED_ENDPOINTS:
                self.server.register_endpoint(
                    ep, RequestType.GET | RequestType.POST, self.request,
                    is_remote=True
                )

    async def _request_initial_subscriptions(self) -> None:
        try:
            await self.klippy_apis.subscribe_objects({'webhooks': None})
        except ServerError:
            logging.exception("Unable to subscribe to webhooks object")
        else:
            logging.info("Webhooks Subscribed")
        try:
            await self.klippy_apis.subscribe_gcode_output()
        except ServerError:
            logging.exception(
                "Unable to register gcode output subscription"
            )
        else:
            logging.info("GCode Output Subscribed")

    async def _check_ready(self) -> None:
        send_id = not self._klippy_identified
        result: Dict[str, Any]
        try:
            result = await self.klippy_apis.get_klippy_info(send_id)
        except ServerError as e:
            if self.init_attempts % LOG_ATTEMPT_INTERVAL == 0 and \
                    self.init_attempts <= MAX_LOG_ATTEMPTS:
                logging.info(
                    f"{e}\nKlippy info request error.  This indicates that\n"
                    f"Klippy may have experienced an error during startup.\n"
                    f"Please check klippy.log for more information")
            return
        version = result.get("software_version", "")
        if version != self._klipper_version:
            self._klipper_version = version
            msg = f"Klipper Version: {version}"
            self.server.add_log_rollover_item("klipper_version", msg)
        klipper_pid: Optional[int] = result.get("process_id")
        if klipper_pid is not None:
            cur_pid: Optional[int] = self._peer_cred.get("process_id")
            if cur_pid is None or klipper_pid != cur_pid:
                self._peer_cred = dict(
                    process_id=klipper_pid,
                    group_id=result.get("group_id", -1),
                    user_id=result.get("user_id", -1)
                )
                await self._get_service_info(klipper_pid)
        self._klippy_info = dict(result)
        state_message: str = self._state.message
        if "state_message" in self._klippy_info:
            state_message = self._klippy_info["state_message"]
            self._state.set_message(state_message)
        if "state" not in result:
            return
        if send_id:
            self._klippy_identified = True
            self._save_path_info()
            await self.server.send_event("server:klippy_identified")
            # Request initial endpoints to register info, emergency stop APIs
            await self._request_endpoints()
        self._state = KlippyState.from_string(result["state"], state_message)
        if self._state != KlippyState.STARTUP:
            await self._request_initial_subscriptions()
            # Register remaining endpoints available
            await self._request_endpoints()
            startup_state = self._state
            await self.server.send_event("server:klippy_started", startup_state)
            self._klippy_started = True
            if self._state != KlippyState.READY:
                logging.info("\n" + self._state.message)
                if (
                    self._state == KlippyState.SHUTDOWN and
                    startup_state != KlippyState.SHUTDOWN
                ):
                    # Klippy shutdown during startup event
                    self.server.send_event("server:klippy_shutdown")
            else:
                await self._verify_klippy_requirements()
                # register methods with klippy
                for method in self.klippy_reg_methods:
                    try:
                        await self.klippy_apis.register_method(method)
                    except ServerError:
                        logging.exception(
                            f"Unable to register method '{method}'")
                self._methods_registered = True
                if self._state == KlippyState.READY:
                    logging.info("Klippy ready")
                    await self.server.send_event("server:klippy_ready")
                    if self._state == KlippyState.SHUTDOWN:
                        # Klippy shutdown during ready event
                        self.server.send_event("server:klippy_shutdown")
                else:
                    logging.info(
                        "Klippy state transition from ready during init, "
                        f"new state: {self._state}"
                    )
            self._klippy_initializing = False

    async def _verify_klippy_requirements(self) -> None:
        result = await self.klippy_apis.get_object_list(default=None)
        if result is None:
            logging.info("Unable to retrieve Klipper Object List")
            return
        req_objs = set(["virtual_sdcard", "display_status", "pause_resume"])
        self._missing_reqs = req_objs - set(result)
        if self._missing_reqs:
            err_str = ", ".join([f"[{o}]" for o in self._missing_reqs])
            logging.info(
                f"\nWarning, unable to detect the following printer "
                f"objects:\n{err_str}\nPlease add the the above sections "
                f"to printer.cfg for full Moonraker functionality.")
        if "virtual_sdcard" not in self._missing_reqs:
            # Update the gcode path
            query_res = await self.klippy_apis.query_objects(
                {'configfile': None}, default=None)
            if query_res is None:
                logging.info("Unable to set SD Card path")
            else:
                config = query_res.get('configfile', {}).get('config', {})
                vsd_config = config.get('virtual_sdcard', {})
                vsd_path = vsd_config.get('path', None)
                if vsd_path is not None:
                    file_manager: FileManager = self.server.lookup_component(
                        'file_manager')
                    file_manager.validate_gcode_path(vsd_path)
                else:
                    logging.info(
                        "Configuration for [virtual_sdcard] not found,"
                        " unable to set SD Card path")

    def _process_command(self, cmd: Dict[str, Any]) -> None:
        method = cmd.get('method', None)
        if method is not None:
            # This is a remote method called from klippy
            if method in self.remote_methods:
                params = cmd.get('params', {})
                self.event_loop.register_callback(
                    self._execute_method, method, **params)
            else:
                logging.info(f"Unknown method received: {method}")
            return
        # This is a response to a request, process
        req_id = cmd.get('id', None)
        request: Optional[KlippyRequest]
        request = self.pending_requests.pop(req_id, None)
        if request is None:
            logging.info(
                f"No request matching request ID: {req_id}, "
                f"response: {cmd}")
            return
        if 'result' in cmd:
            result = cmd['result']
            if not result:
                result = "ok"
            request.set_result(result)
        else:
            err: Union[str, Dict[str, str]]
            err = cmd.get('error', "Malformed Klippy Response")
            if isinstance(err, dict):
                err = err.get("message", "Malformed Klippy Response")
            request.set_exception(ServerError(err, 400))

    async def _execute_method(self, method_name: str, **kwargs) -> None:
        try:
            ret = self.remote_methods[method_name](**kwargs)
            if ret is not None:
                await ret
        except Exception:
            logging.exception(f"Error running remote method: {method_name}")

    def _process_gcode_response(self, response: str) -> None:
        self.server.send_event("server:gcode_response", response)

    def _process_status_update(
        self, eventtime: float, status: Dict[str, Dict[str, Any]]
    ) -> None:
        for field, item in status.items():
            self.subscription_cache.setdefault(field, {}).update(item)
        if 'webhooks' in status:
            wh: Dict[str, str] = status['webhooks']
            state_message: str = self._state.message
            if "state_message" in wh:
                state_message = wh["state_message"]
                self._state.set_message(state_message)
            # XXX - process other states (startup, ready, error, etc)?
            if "state" in wh:
                new_state = KlippyState.from_string(wh["state"], state_message)
                if (
                    new_state == KlippyState.SHUTDOWN and
                    not self._klippy_initializing and
                    self._state != KlippyState.SHUTDOWN
                ):
                    # If the shutdown state is received during initialization
                    # defer the event, the init routine will handle it.
                    logging.info("Klippy has shutdown")
                    self.server.send_event("server:klippy_shutdown")
                self._state = new_state
        for conn, sub in self.subscriptions.items():
            conn_status: Dict[str, Any] = {}
            for name, fields in sub.items():
                if name in status:
                    val: Dict[str, Any] = dict(status[name])
                    if fields is not None:
                        val = {k: v for k, v in val.items() if k in fields}
                    if val:
                        conn_status[name] = val
            conn.send_status(conn_status, eventtime)

    async def request(self, web_request: WebRequest) -> Any:
        if not self.is_connected():
            raise ServerError("Klippy Host not connected", 503)
        rpc_method = web_request.get_endpoint()
        if rpc_method == "objects/subscribe":
            return await self._request_subscripton(web_request)
        else:
            if rpc_method == "gcode/script":
                script = web_request.get_str('script', "")
                if script:
                    self.server.send_event(
                        "klippy_connection:gcode_received", script)
            return await self._request_standard(web_request)

    async def _request_subscripton(self, web_request: WebRequest) -> Dict[str, Any]:
        async with self.subscription_lock:
            args = web_request.get_args()
            conn = web_request.get_subscribable()
            if conn is None:
                raise self.server.error(
                    "No connection associated with subscription request"
                )
            requested_sub: Subscription = args.get('objects', {})
            all_subs: Subscription = dict(requested_sub)
            # Build the subscription request from a superset of all client subscriptions
            for sub in self.subscriptions.values():
                for obj, items in sub.items():
                    if obj in all_subs:
                        prev_items = all_subs[obj]
                        if items is None or prev_items is None:
                            all_subs[obj] = None
                        else:
                            uitems = list(set(prev_items) | set(items))
                            all_subs[obj] = uitems
                    else:
                        all_subs[obj] = items
            args['objects'] = all_subs
            args['response_template'] = {'method': "process_status_update"}

            result = await self._request_standard(web_request, 20.0)

            # prune the status response
            pruned_status: Dict[str, Dict[str, Any]] = {}
            status_diff: Dict[str, Dict[str, Any]] = {}
            all_status: Dict[str, Dict[str, Any]] = result['status']
            for obj, fields in all_status.items():
                # Diff the current cache, then update the cache
                if obj in self.subscription_cache:
                    cached_status = self.subscription_cache[obj]
                    for field_name, value in fields.items():
                        if field_name not in cached_status:
                            continue
                        if value != cached_status[field_name]:
                            status_diff.setdefault(obj, {})[field_name] = value
                if obj in CACHE_EXCLUSIONS:
                    # Make a shallow copy so we can pop off fields we want to
                    # exclude from the cache without modifying the return value
                    fields_to_cache = dict(fields)
                    removed: List[str] = []
                    for excluded_field in CACHE_EXCLUSIONS[obj]:
                        if excluded_field in fields_to_cache:
                            removed.append(excluded_field)
                            del fields_to_cache[excluded_field]
                    if removed:
                        logging.debug(
                            "Removed excluded fields from subscription cache: "
                            f"{obj}: {removed}"
                        )
                    self.subscription_cache[obj] = fields_to_cache
                else:
                    self.subscription_cache[obj] = fields
                # Prune Response
                if obj in requested_sub:
                    valid_fields = requested_sub[obj]
                    if valid_fields is None:
                        pruned_status[obj] = fields
                    else:
                        pruned_status[obj] = {
                            k: v for k, v in fields.items() if k in valid_fields
                        }
            if status_diff:
                # The response to the status request contains changed data, so it
                # is necessary to manually push the status update to existing
                # subscribers
                logging.debug(
                    f"Detected status difference during subscription: {status_diff}"
                )
                self._process_status_update(result["eventtime"], status_diff)
            for obj_name in list(self.subscription_cache.keys()):
                # Prune the cache to match the current status response
                if obj_name not in all_status:
                    del self.subscription_cache[obj_name]
            result['status'] = pruned_status
            self.subscriptions[conn] = requested_sub
            return result

    async def _request_standard(
        self, web_request: WebRequest, timeout: Optional[float] = None
    ) -> Any:
        rpc_method = web_request.get_endpoint()
        args = web_request.get_args()
        # Create a base klippy request
        base_request = KlippyRequest(rpc_method, args)
        self.pending_requests[base_request.id] = base_request
        self.event_loop.register_callback(self._write_request, base_request)
        try:
            return await base_request.wait(timeout)
        finally:
            self.pending_requests.pop(base_request.id, None)

    def remove_subscription(self, conn: APITransport) -> None:
        self.subscriptions.pop(conn, None)

    def is_connected(self) -> bool:
        return self.writer is not None and not self.closing

    def is_ready(self) -> bool:
        return self._state == KlippyState.READY

    def is_printing(self) -> bool:
        if not self.is_ready():
            return False
        job_state: JobState = self.server.lookup_component("job_state")
        stats = job_state.get_last_stats()
        return stats.get("state", "") == "printing"

    def get_subscription_cache(self) -> Dict[str, Dict[str, Any]]:
        return self.subscription_cache

    async def rollover_log(self) -> None:
        if "unit_name" not in self._service_info:
            raise self.server.error(
                "Unable to detect Klipper Service, cannot perform "
                "manual rollover"
            )
        logfile: Optional[str] = self._klippy_info.get("log_file", None)
        if logfile is None:
            raise self.server.error(
                "Unable to detect path to Klipper log file"
            )
        if self.is_printing():
            raise self.server.error("Cannot rollover log while printing")
        logpath = pathlib.Path(logfile).expanduser().resolve()
        if not logpath.is_file():
            raise self.server.error(
                f"No file at {logpath} exists, cannot perform rollover"
            )
        machine: Machine = self.server.lookup_component("machine")
        await machine.do_service_action("stop", self.unit_name)
        suffix = time.strftime("%Y-%m-%d_%H-%M-%S", time.localtime())
        new_path = pathlib.Path(f"{logpath}.{suffix}")

        def _do_file_op() -> None:
            if new_path.exists():
                new_path.unlink()
            logpath.rename(new_path)

        await self.event_loop.run_in_thread(_do_file_op)
        await machine.do_service_action("start", self.unit_name)

    async def _on_connection_closed(self) -> None:
        self._klippy_identified = False
        self._klippy_initializing = False
        self._klippy_started = False
        self._methods_registered = False
        self._state = KlippyState.DISCONNECTED
        self._state.set_message("Klippy Disconnected")
        for request in self.pending_requests.values():
            request.set_exception(ServerError("Klippy Disconnected", 503))
        self.pending_requests = {}
        self.subscriptions = {}
        self.subscription_cache.clear()
        self._peer_cred = {}
        self._missing_reqs.clear()
        logging.info("Klippy Connection Removed")
        await self.server.send_event("server:klippy_disconnect")
        if self.server.is_running():
            # Reconnect if server is running
            loop = self.event_loop
            self.connection_task = loop.create_task(self._do_connect())

    async def close(self, wait_closed: bool = False) -> None:
        if self.closing:
            if wait_closed:
                await self.connection_mutex.acquire()
                self.connection_mutex.release()
            return
        self.closing = True
        if (
            self.connection_task is not None and
            not self.connection_task.done()
        ):
            self.connection_task.cancel()
        async with self.connection_mutex:
            if self.writer is not None:
                try:
                    self.writer.close()
                    await self.writer.wait_closed()
                except Exception:
                    logging.exception("Error closing Klippy Unix Socket")
                self.writer = None
                await self._on_connection_closed()
        self.closing = False

# Basic KlippyRequest class, easily converted to dict for json encoding
class KlippyRequest:
    def __init__(self, rpc_method: str, params: Dict[str, Any]) -> None:
        self.id = id(self)
        self.rpc_method = rpc_method
        self.params = params
        self._fut = asyncio.get_running_loop().create_future()

    async def wait(self, timeout: Optional[float] = None) -> Any:
        start_time = time.time()
        to = timeout or 60.
        while True:
            try:
                return await asyncio.wait_for(asyncio.shield(self._fut), to)
            except asyncio.TimeoutError:
                if timeout is not None:
                    self._fut.cancel()
                    raise ServerError("Klippy request timed out", 500) from None
                pending_time = time.time() - start_time
                logging.info(
                    f"Request '{self.rpc_method}' pending: "
                    f"{pending_time:.2f} seconds"
                )

    def set_exception(self, exc: Exception) -> None:
        if not self._fut.done():
            self._fut.set_exception(exc)

    def set_result(self, result: Any) -> None:
        if not self._fut.done():
            self._fut.set_result(result)

    def to_dict(self) -> Dict[str, Any]:
        return {
            'id': self.id,
            'method': self.rpc_method,
            'params': self.params
        }

def load_component(config: ConfigHelper) -> KlippyConnection:
    return KlippyConnection(config)



#!/usr/bin/python3
# Builds zip release files for Moonraker and Klipper

import os
import sys
import argparse
import shutil
import tempfile
import json
import pathlib
import time
import traceback
import subprocess
import re
from typing import Dict, Any, List, Set, Optional

MOONRAKER_PATH = os.path.abspath(
    os.path.join(os.path.dirname(__file__), ".."))
sys.path.append(os.path.join(MOONRAKER_PATH, "moonraker"))
from utils import hash_directory, retrieve_git_version  # noqa:E402

# Dirs and exts to ignore when calculating the repo hash
IGNORE_DIRS = ["out", "lib", "test", "docs", "__pycache__"]
IGNORE_EXTS = [".o", ".so", ".pyc", ".pyo", ".pyd", ".yml", ".yaml"]

# Files not to include in the source package
SKIP_FILES = [".gitignore", ".gitattributes", ".readthedocs.yaml",
              "mkdocs.yml", "__pycache__"]

RELEASE_URL = "https://api.github.com/repos/Arksine/moonraker/releases"
GIT_MAX_LOG_CNT = 100
GIT_LOG_FMT = \
    "sha:%H%x1Dauthor:%an%x1Ddate:%ct%x1Dsubject:%s%x1Dmessage:%b%x1E"
OWNER_REPOS = {
    'moonraker': "arksine/moonraker",
    'klippy': "klipper3d/klipper"
}
INSTALL_SCRIPTS = {
    'klippy': {
        'debian': "install-octopi.sh",
        'arch': "install-arch.sh",
        'centos': "install-centos.sh"
    },
    'moonraker': {
        'debian': "install-moonraker.sh"
    }
}

class CopyIgnore:
    def __init__(self, root_dir: str) -> None:
        self.root_dir = root_dir

    def __call__(self, dir_path: str, dir_items: List[str]) -> List[str]:
        ignored: List[str] = []
        for item in dir_items:
            if item in SKIP_FILES:
                ignored.append(item)
            elif dir_path == self.root_dir:
                full_path = os.path.join(dir_path, item)
                # Ignore all hidden directories in the root
                if os.path.isdir(full_path) and item[0] == ".":
                    ignored.append(item)
        return ignored

def search_install_script(data: str,
                          regex: str,
                          exclude: str
                          ) -> List[str]:
    items: Set[str] = set()
    lines: List[str] = re.findall(regex, data)
    for line in lines:
        items.update(line.strip().split())
    try:
        items.remove(exclude)
    except KeyError:
        pass
    return list(items)

def generate_dependency_info(repo_path: str, app_name: str) -> None:
    inst_scripts = INSTALL_SCRIPTS[app_name]
    package_info: Dict[str, Any] = {}
    for distro, script_name in inst_scripts.items():
        script_path = os.path.join(repo_path, "scripts", script_name)
        script = pathlib.Path(script_path)
        if not script.exists():
            continue
        data = script.read_text()
        packages: List[str] = search_install_script(
            data, r'PKGLIST="(.*)"', "${PKGLIST}")
        package_info[distro] = {'packages': sorted(packages)}
        if distro == "arch":
            aur_packages: List[str] = search_install_script(
                data, r'AURLIST="(.*)"', "${AURLIST}")
            package_info[distro]['aur_packages'] = sorted(aur_packages)
    req_file_name = os.path.join(repo_path, "scripts",
                                 f"{app_name}-requirements.txt")
    req_file = pathlib.Path(req_file_name)
    python_reqs: List[str] = []
    if req_file.exists():
        req_data = req_file.read_text()
        lines = [line.strip() for line in req_data.split('\n')
                 if line.strip()]
        for line in lines:
            comment_idx = line.find('#')
            if comment_idx == 0:
                continue
            if comment_idx > 0:
                line = line[:comment_idx].strip()
            python_reqs.append(line)
    package_info['python'] = sorted(python_reqs)
    dep_file = pathlib.Path(os.path.join(repo_path, ".dependencies"))
    dep_file.write_text(json.dumps(package_info))

def clean_repo(path: str) -> None:
    # Obtain version info from "git" program
    prog = ('git', '-C', path, 'clean', '-x', '-f', '-d')
    process = subprocess.Popen(prog, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, cwd=path)
    retcode = process.wait()
    if retcode != 0:
        print(f"Error running git clean: {path}")

def get_releases() -> List[Dict[str, Any]]:
    print("Fetching Release List...")
    prog = ('curl', '-H', "Accept: application/vnd.github.v3+json",
            RELEASE_URL)
    process = subprocess.Popen(prog, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    response, err = process.communicate()
    retcode = process.wait()
    if retcode != 0:
        print(f"Release list request returned with code {retcode},"
              f" response:\n{err.decode()}")
        return []
    releases = json.loads(response.decode().strip())
    print(f"Found {len(releases)} releases")
    return releases

def get_last_release_info(moonraker_version: str,
                          is_beta: bool,
                          releases: List[Dict[str, Any]]
                          ) -> Dict[str, Any]:
    print("Searching for previous release assets...")
    cur_tag, commit_count = moonraker_version.split('-', 2)[:2]
    release_assets = []
    matched_tag: Optional[str] = None
    for release in releases:
        if int(commit_count) != 0:
            # This is build is not being done against a fresh release,
            # return release info from a matching tag
            if release['tag_name'] == cur_tag:
                release_assets = release['assets']
                matched_tag = cur_tag
                break
        else:
            # Get the most recent non-matching tag
            if release['tag_name'] == cur_tag:
                continue
            if is_beta or not release['prerelease']:
                # Get the last tagged release.  If we are building a beta,
                # that is the most recent release.  Otherwise we should
                # omit pre-releases
                release_assets = release['assets']
                matched_tag = release['tag_name']
                break
    if matched_tag is None:
        print("No matching release found")
        matched_tag = "No Tag"
    else:
        print(f"Found release: {matched_tag}")

    asset_url: Optional[str] = None
    content_type: str = ""
    for asset in release_assets:
        if asset['name'] == "RELEASE_INFO":
            asset_url = asset['browser_download_url']
            content_type = asset['content_type']
            break
    if asset_url is None:
        print(f"RELEASE_INFO asset not found in release: {matched_tag}")
        return {}
        # This build is prior to a tagged release, so fetch the current tag
    print(f"Release Info Download URL: {asset_url}")
    prog = ('curl', '-L', '-H', f"Accept: {content_type}", asset_url)
    process = subprocess.Popen(prog, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE)
    response, err = process.communicate()
    retcode = process.wait()
    if retcode != 0:
        print("Request for release info failed")
        return {}
    resp = response.decode().strip()
    print(f"Found Info for release {matched_tag}")
    return json.loads(resp)

def get_commit_log(path: str,
                   release_info: Dict[str, Any]
                   ) -> List[Dict[str, Any]]:
    print(f"Preparing commit log for {path.split('/')[-1]}")
    start_sha = release_info.get('commit_hash', None)
    prog = ['git', '-C', path, 'log', f'--format={GIT_LOG_FMT}',
            f'--max-count={GIT_MAX_LOG_CNT}']
    if start_sha is not None:
        prog = ['git', '-C', path, 'log', f'{start_sha}..HEAD',
                f'--format={GIT_LOG_FMT}', f'--max-count={GIT_MAX_LOG_CNT}']
    process = subprocess.Popen(prog, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, cwd=path)
    response, err = process.communicate()
    retcode = process.wait()
    if retcode != 0:
        return []
    resp = response.decode().strip()
    commit_log: List[Dict[str, Any]] = []
    for log_entry in resp.split('\x1E'):
        log_entry = log_entry.strip()
        if not log_entry:
            continue
        log_items = [li.strip() for li in log_entry.split('\x1D')
                     if li.strip()]
        cbh = [li.split(':', 1) for li in log_items]
        commit_log.append(dict(cbh))  # type: ignore
    print(f"Found {len(commit_log)} commits")
    return commit_log

def get_commit_hash(path: str) -> str:
    prog = ('git', '-C', path, 'rev-parse', 'HEAD')
    process = subprocess.Popen(prog, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, cwd=path)
    commit_hash, err = process.communicate()
    retcode = process.wait()
    if retcode == 0:
        return commit_hash.strip().decode()
    raise Exception(f"Failed to get commit hash: {commit_hash.decode()}")

def generate_version_info(path: str,
                          source_dir: str,
                          channel: str,
                          release_tag: Optional[str] = None
                          ) -> Dict[str, Any]:
    print(f"Generating version info: {source_dir}")
    clean_repo(path)
    owner_repo = OWNER_REPOS[source_dir]
    curtime = int(time.time())
    date_str = time.strftime("%Y%m%d", time.gmtime(curtime))
    version = retrieve_git_version(path)
    if release_tag is None:
        release_tag = version.split('-')[0]
    source_hash = hash_directory(path, IGNORE_EXTS, IGNORE_DIRS)
    long_version = f"{version}-moonraker-{date_str}"
    release_info = {
        'git_version': version,
        'long_version': long_version,
        'commit_hash': get_commit_hash(path),
        'source_checksum': source_hash,
        'ignored_exts': IGNORE_EXTS,
        'ignored_dirs': IGNORE_DIRS,
        'build_date': curtime,
        'channel': channel,
        'owner_repo': owner_repo,
        'host_repo': OWNER_REPOS['moonraker'],
        'release_tag': release_tag
    }
    vfile = pathlib.Path(os.path.join(path, source_dir, ".version"))
    vfile.write_text(long_version)
    rfile = pathlib.Path(os.path.join(path, ".release_info"))
    rfile.write_text(json.dumps(release_info))
    generate_dependency_info(path, source_dir)
    return release_info

def create_zip(repo_path: str,
               repo_name: str,
               output_path: str
               ) -> None:
    print(f"Creating Zip Release: {repo_name}")
    zip_path = os.path.join(output_path, repo_name)
    with tempfile.TemporaryDirectory() as tmp_dir:
        dest_path = os.path.join(tmp_dir, repo_name)
        ingore_cb = CopyIgnore(repo_path)
        shutil.copytree(repo_path, dest_path, ignore=ingore_cb)
        shutil.make_archive(zip_path, "zip", root_dir=dest_path)

def main() -> None:
    # Parse start arguments
    parser = argparse.ArgumentParser(
        description="Generates zip releases for Moonraker and Klipper")
    parser.add_argument(
        "-k", "--klipper", default="~/klipper",
        metavar='<klipper_path>',
        help="Path to Klipper git repo")
    parser.add_argument(
        "-o", "--output", default=os.path.join(MOONRAKER_PATH, ".dist"),
        metavar='<output_path>', help="Path to output directory")
    parser.add_argument(
        "-b", "--beta", action='store_true',
        help="Tag release as beta")
    args = parser.parse_args()
    kpath: str = os.path.abspath(os.path.expanduser(args.klipper))
    opath: str = os.path.abspath(os.path.expanduser(args.output))
    is_beta: bool = args.beta
    channel = "beta" if is_beta else "stable"
    if not os.path.exists(kpath):
        print(f"Invalid path to Klipper: {kpath}")
        sys.exit(-1)
    if not os.path.exists(opath):
        print(f"Invalid output path: {opath}")
        sys.exit(-1)
    releases = get_releases()
    all_info: Dict[str, Dict[str, Any]] = {}
    try:
        print("Generating Moonraker Zip Distribution...")
        all_info['moonraker'] = generate_version_info(
            MOONRAKER_PATH, "moonraker", channel)
        create_zip(MOONRAKER_PATH, 'moonraker', opath)
        rtag: str = all_info['moonraker']['release_tag']
        print("Generating Klipper Zip Distribution...")
        all_info['klipper'] = generate_version_info(
            kpath, "klippy", channel, rtag)
        create_zip(kpath, 'klipper', opath)
        info_file = pathlib.Path(os.path.join(opath, "RELEASE_INFO"))
        info_file.write_text(json.dumps(all_info))
        last_rinfo = get_last_release_info(
            all_info['moonraker']['git_version'], is_beta, releases)
        commit_log = {}
        commit_log['moonraker'] = get_commit_log(
            MOONRAKER_PATH, last_rinfo.get('moonraker', {}))
        commit_log['klipper'] = get_commit_log(
            kpath, last_rinfo.get('klipper', {}))
        clog_file = pathlib.Path(os.path.join(opath, "COMMIT_LOG"))
        clog_file.write_text(json.dumps(commit_log))
    except Exception:
        print("Error Creating Zip Distribution")
        traceback.print_exc(file=sys.stdout)
        sys.exit(-1)
    print(f"Build Complete.  Files are located at '{opath}'")


if __name__ == "__main__":
    main()



# Print Job Queue Implementation
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import asyncio
import time
import logging
from ..common import JobEvent, RequestType

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Dict,
    List,
    Union,
)
if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest, UserInfo
    from .klippy_apis import KlippyAPI
    from .file_manager.file_manager import FileManager
    from .job_state import JobState

class JobQueue:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.queued_jobs: Dict[str, QueuedJob] = {}
        self.lock = asyncio.Lock()
        self.pause_requested: bool = False
        self.load_on_start = config.getboolean("load_on_startup", False)
        self.automatic = config.getboolean("automatic_transition", False)
        self.queue_state: str = "paused"
        self.job_delay = config.getfloat("job_transition_delay", 0.01)
        if self.job_delay <= 0.:
            raise config.error(
                "Value for option 'job_transition_delay' in section [job_queue]"
                " must be above 0.0")
        self.job_transition_gcode = config.get(
            "job_transition_gcode", "").strip()
        self.pop_queue_handle: Optional[asyncio.TimerHandle] = None

        self.server.register_event_handler(
            "server:klippy_ready", self._handle_ready)
        self.server.register_event_handler(
            "server:klippy_shutdown", self._handle_shutdown)
        self.server.register_event_handler(
            "job_state:state_changed", self._on_job_state_changed
        )

        self.server.register_notification("job_queue:job_queue_changed")
        self.server.register_remote_method("pause_job_queue", self.pause_queue)
        self.server.register_remote_method("start_job_queue", self.start_queue)

        self.server.register_endpoint(
            "/server/job_queue/job", RequestType.POST | RequestType.DELETE,
            self._handle_job_request
        )
        self.server.register_endpoint(
            "/server/job_queue/pause", RequestType.POST, self._handle_pause_queue
        )
        self.server.register_endpoint(
            "/server/job_queue/start", RequestType.POST, self._handle_start_queue
        )
        self.server.register_endpoint(
            "/server/job_queue/status", RequestType.GET, self._handle_queue_status
        )
        self.server.register_endpoint(
            "/server/job_queue/jump", RequestType.POST, self._handle_jump
        )

    async def _handle_ready(self) -> None:
        async with self.lock:
            if not self.load_on_start or not self.queued_jobs:
                return
            # start a queued print
            if self.queue_state in ['ready', 'paused']:
                event_loop = self.server.get_event_loop()
                self._set_queue_state("loading")
                self.pop_queue_handle = event_loop.delay_callback(
                    1., self._pop_job, False)

    async def _handle_shutdown(self) -> None:
        has_requested_pause = self.pause_requested
        await self.pause_queue()
        self.pause_requested = has_requested_pause

    async def _on_job_state_changed(self, job_event: JobEvent, *args) -> None:
        if job_event == JobEvent.COMPLETE:
            await self._on_job_complete()
        elif job_event.aborted:
            await self._on_job_abort()

    async def _on_job_complete(self) -> None:
        if not self.automatic:
            return
        async with self.lock:
            # Transition to the next job in the queue
            if self.queue_state == "ready" and self.queued_jobs:
                event_loop = self.server.get_event_loop()
                self._set_queue_state("loading")
                self.pop_queue_handle = event_loop.delay_callback(
                    self.job_delay, self._pop_job)

    async def _on_job_abort(self) -> None:
        async with self.lock:
            if self.queued_jobs:
                self._set_queue_state("paused")

    async def _pop_job(self, need_transition: bool = True) -> None:
        self.pop_queue_handle = None
        async with self.lock:
            if self.queue_state == "paused":
                return
            if not self.queued_jobs:
                self._set_queue_state("paused")
                return
            kapis: KlippyAPI = self.server.lookup_component('klippy_apis')
            uid, job = list(self.queued_jobs.items())[0]
            filename = str(job)
            can_print = await self._check_can_print()
            if not can_print or self.queue_state != "loading":
                self._set_queue_state("paused")
                return
            try:
                if self.job_transition_gcode and need_transition:
                    await kapis.run_gcode(self.job_transition_gcode)
                    # Check to see if the queue was paused while running
                    # the job transition gcode
                    if self.queue_state != "loading":
                        self._set_queue_state("paused")
                        raise self.server.error(
                            "Queue State Changed during Transition Gcode")
                self._set_queue_state("starting")
                await kapis.start_print(
                    filename, wait_klippy_started=True, user=job.user
                )
            except self.server.error:
                logging.exception(f"Error Loading print: {filename}")
                self._set_queue_state("paused")
            else:
                self.queued_jobs.pop(uid, None)
                if self.queue_state == "starting":
                    # Set the queue to ready if items are in the queue
                    # and auto transition is enabled
                    set_ready = len(self.queued_jobs) > 0 and self.automatic
                    self.queue_state = "ready" if set_ready else "paused"
                self._send_queue_event(action="job_loaded")

    async def _check_can_print(self) -> bool:
        # Query the latest stats
        kapis: KlippyAPI = self.server.lookup_component('klippy_apis')
        try:
            result = await kapis.query_objects({"print_stats": None})
        except Exception:
            # Klippy not connected
            return False
        if 'print_stats' not in result:
            return False
        state: str = result['print_stats']['state']
        if state in ["printing", "paused"]:
            return False
        return True

    async def queue_job(self,
                        filenames: Union[str, List[str]],
                        check_exists: bool = True,
                        reset: bool = False,
                        user: Optional[UserInfo] = None
                        ) -> None:
        async with self.lock:
            # Make sure that the file exists
            if isinstance(filenames, str):
                filenames = [filenames]
            if check_exists:
                # Make sure all files exist before adding them to the queue
                for fname in filenames:
                    self._check_job_file(fname)
            if reset:
                self.queued_jobs.clear()
            for fname in filenames:
                queued_job = QueuedJob(fname, user)
                self.queued_jobs[queued_job.job_id] = queued_job
            self._send_queue_event(action="jobs_added")
            if self.automatic and not self.pause_requested:
                jstate: JobState = self.server.lookup_component("job_state")
                last_evt = jstate.get_last_job_event()
                if last_evt.is_printing or last_evt == JobEvent.PAUSED:
                    self._set_queue_state("ready")

    async def delete_job(self,
                         job_ids: Union[str, List[str]],
                         all: bool = False
                         ) -> None:
        async with self.lock:
            if not self.queued_jobs:
                # No jobs in queue, nothing to delete
                return
            if all:
                self.queued_jobs.clear()
            elif job_ids:
                if isinstance(job_ids, str):
                    job_ids = [job_ids]
                for uid in job_ids:
                    self.queued_jobs.pop(uid, None)
            else:
                # Don't notify, nothing was deleted
                return
            self._send_queue_event(action="jobs_removed")

    async def pause_queue(self) -> None:
        self._set_queue_state("paused")
        if self.pop_queue_handle is not None:
            self.pop_queue_handle.cancel()
            self.pop_queue_handle = None
        # Acquire the lock to wait for any pending operations to complete
        async with self.lock:
            self.pause_requested = True

    async def start_queue(self) -> None:
        async with self.lock:
            self.pause_requested = False
            if self.queue_state in ("ready", "paused"):
                if not self.queued_jobs:
                    self._set_queue_state("paused")
                elif await self._check_can_print():
                    self._set_queue_state("loading")
                    event_loop = self.server.get_event_loop()
                    self.pop_queue_handle = event_loop.delay_callback(
                        0.01, self._pop_job, False
                    )
                else:
                    self._set_queue_state("ready" if self.automatic else "paused")

    def _job_map_to_list(self) -> List[Dict[str, Any]]:
        cur_time = time.time()
        return [job.as_dict(cur_time) for
                job in self.queued_jobs.values()]

    def _check_job_file(self, job_name: str) -> None:
        fm: FileManager = self.server.lookup_component('file_manager')
        if not fm.check_file_exists("gcodes", job_name):
            raise self.server.error(
                f"G-Code File {job_name} does not exist")

    def _set_queue_state(self, new_state: str) -> None:
        if new_state != self.queue_state:
            self.queue_state = new_state
            self._send_queue_event()

    def _send_queue_event(self, action: str = "state_changed"):
        updated_queue: Optional[List[Dict[str, Any]]] = None
        if action != "state_changed":
            updated_queue = self._job_map_to_list()
        event_loop = self.server.get_event_loop()
        event_loop.delay_callback(
            .05, self.server.send_event, "job_queue:job_queue_changed",
            {
                'action': action,
                'updated_queue': updated_queue,
                'queue_state': self.queue_state
            })

    async def _handle_job_request(
        self, web_request: WebRequest
    ) -> Dict[str, Any]:
        req_type = web_request.get_request_type()
        if req_type == RequestType.POST:
            files = web_request.get_list('filenames')
            reset = web_request.get_boolean("reset", False)
            # Validate that all files exist before queueing
            user = web_request.get_current_user()
            await self.queue_job(files, reset=reset, user=user)
        elif req_type == RequestType.DELETE:
            if web_request.get_boolean("all", False):
                await self.delete_job([], all=True)
            else:
                job_ids = web_request.get_list('job_ids')
                await self.delete_job(job_ids)
        else:
            raise self.server.error(f"Invalid request type: {req_type}")
        return {
            'queued_jobs': self._job_map_to_list(),
            'queue_state': self.queue_state
        }

    async def _handle_pause_queue(self,
                                  web_request: WebRequest
                                  ) -> Dict[str, Any]:
        await self.pause_queue()
        return {
            'queued_jobs': self._job_map_to_list(),
            'queue_state': self.queue_state
        }

    async def _handle_start_queue(self,
                                  web_request: WebRequest
                                  ) -> Dict[str, Any]:
        await self.start_queue()
        return {
            'queued_jobs': self._job_map_to_list(),
            'queue_state': self.queue_state
        }

    async def _handle_queue_status(self,
                                   web_request: WebRequest
                                   ) -> Dict[str, Any]:
        return {
            'queued_jobs': self._job_map_to_list(),
            'queue_state': self.queue_state
        }

    async def _handle_jump(self, web_request: WebRequest) -> Dict[str, Any]:
        job_id: str = web_request.get("job_id")
        async with self.lock:
            job = self.queued_jobs.pop(job_id, None)
            if job is None:
                raise self.server.error(f"Invalid job id: {job_id}")
            new_queue = {job_id: job}
            new_queue.update(self.queued_jobs)
            self.queued_jobs = new_queue
        return {
            'queued_jobs': self._job_map_to_list(),
            'queue_state': self.queue_state
        }

    async def close(self):
        await self.pause_queue()

class QueuedJob:
    def __init__(self, filename: str, user: Optional[UserInfo] = None) -> None:
        self.filename = filename
        self.job_id = f"{id(self):016X}"
        self.time_added = time.time()
        self._user = user

    def __str__(self) -> str:
        return self.filename

    @property
    def user(self) -> Optional[UserInfo]:
        return self._user

    def as_dict(self, cur_time: float) -> Dict[str, Any]:
        return {
            'filename': self.filename,
            'job_id': self.job_id,
            'time_added': self.time_added,
            'time_in_queue': cur_time - self.time_added
        }

def load_component(config: ConfigHelper) -> JobQueue:
    return JobQueue(config)



# Python Package Update Deployment
#
# Copyright (C) 2024  Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.

from __future__ import annotations
import asyncio
import re
import logging
from enum import Enum
from ...utils.source_info import normalize_project_name, load_distribution_info
from ...utils.versions import PyVersion, GitVersion
from ...utils import pip_utils, json_wrapper
from .app_deploy import AppDeploy, Channel, DISTRO_ALIASES

# Annotation imports
from typing import (
    TYPE_CHECKING,
    Any,
    Optional,
    Dict,
    List,
    cast
)

if TYPE_CHECKING:
    from ...confighelper import ConfigHelper
    from ...utils.source_info import PackageInfo
    from ...components.file_manager.file_manager import FileManager
    from .update_manager import CommandHelper

class PackageSource(Enum):
    PIP = 0
    GITHUB = 1
    UNKNOWN = 2


class PythonDeploy(AppDeploy):
    def __init__(self, config: ConfigHelper, cmd_helper: CommandHelper) -> None:
        super().__init__(config, cmd_helper, "Python Package")
        self._configure_virtualenv(config)
        if self.virtualenv is None:
            raise config.error(
                f"[{config.get_name()}]: Option 'virtualenv' must specify a valid "
                "the path to a Python virtualenv"
            )
        fm: FileManager = self.server.lookup_component("file_manager")
        fm.add_reserved_path(f"update_manager {self.name}", self.virtualenv)
        self._configure_managed_services(config)
        self.primary_branch = config.get("primary_branch", None)
        self.project_name = config.get("project_name", self.name)
        self.source: PackageSource = PackageSource.UNKNOWN
        self.repo_url: str = ""
        self.repo_owner: str = "?"
        self.repo_name: str = "?"
        self.current_version: PyVersion = PyVersion("?")
        self.git_version: GitVersion = GitVersion("?")
        self.current_sha: str = "?"
        self.upstream_version: PyVersion = self.current_version
        self.upstream_sha: str = "?"
        self.rollback_ref: str = "?"
        self.warnings: List[str] = []
        package_info = load_distribution_info(self.virtualenv, self.project_name)
        self._detect_update_source(package_info)
        self._update_current_version(package_info)
        self.changelog: str = self._get_url(["changelog"], package_info)
        self.system_deps = self._parse_system_dependencies(package_info)
        self._is_valid = len(self.warnings) == 0

    async def initialize(self) -> Dict[str, Any]:
        storage = await super().initialize()
        self.upstream_sha = storage.get("upstream_commit", "?")
        self.upstream_version = PyVersion(storage.get("upstream_version", "?"))
        self.rollback_ref = storage.get("rollback_ref", "?")
        if not self.needs_refresh():
            self._log_package_info()
        return storage

    def get_persistent_data(self) -> Dict[str, Any]:
        storage = super().get_persistent_data()
        storage["upstream_commit"] = self.upstream_sha
        storage["upstream_version"] = self.upstream_version.full_version
        storage["rollback_ref"] = self.rollback_ref
        return storage

    def get_update_status(self) -> Dict[str, Any]:
        status = super().get_update_status()
        status.update({
            "detected_type": "python_package",
            "branch": self.primary_branch,
            "owner": self.repo_owner,
            "repo_name": self.repo_name,
            "version": self.current_version.short_version,
            "remote_version": self.upstream_version.short_version,
            "current_hash": self.current_sha,
            "remote_hash": self.upstream_sha,
            "is_dirty": self.git_version.dirty,
            "changelog_url": self.changelog,
            "full_version_string": self.current_version.full_version,
            "pristine": not self.git_version.dirty,
            "warnings": self.warnings
        })
        return status

    def _add_warning(self, msg: str) -> None:
        self.warnings.append(msg)
        self.log_info(msg)

    def _detect_update_source(self, package_info: PackageInfo) -> None:
        self.source = PackageSource.UNKNOWN
        direct_url_data = package_info.direct_url_data
        if direct_url_data is None:
            self.source = PackageSource.PIP
            self.repo_url = self._get_url(["repository", "repo"], package_info)
            self._match_repo_url()
            return
        self.log_debug(f"Direct URL info: {direct_url_data}")
        vcs_info: Dict[str, str] = direct_url_data.get("vcs_info", {})
        if vcs_info.get("vcs", "") != "git":
            self._add_warning(
                "Package installed from source other than pypi or git: "
                f"{direct_url_data}"
            )
            return
        try:
            self.current_sha = vcs_info["commit_id"]
            self.repo_url = direct_url_data["url"]
        except KeyError:
            self._add_warning("Failed to retrive direct_url vcs info")
            return
        if not self._match_repo_url():
            self._add_warning(f"Invalid repo url: {self.repo_url}")
            return
        self.source = PackageSource.GITHUB

    def _match_repo_url(self) -> bool:
        url_match = re.match(
            r"https://(?:www\.)?github\.com/(?P<owner>.+?)/(?P<proj>.+?)(?:\.git|$)",
            self.repo_url, re.IGNORECASE
        )
        if url_match is None:
            return False
        self.repo_owner = url_match["owner"] or "?"
        self.repo_name = url_match["proj"] or "?"
        return True

    def _get_url(self, keys: List[str], package_info: PackageInfo) -> str:
        release_info = package_info.release_info
        primary = keys[0]
        if release_info is not None:
            urls: Dict[str, Any] = release_info.get("urls", {})
            for name, url in urls.items():
                if name.lower() in keys:
                    return url
            self.log_debug(f"Unable to find {primary} url in release info")
        # Fallback to Metadata
        metadata = package_info.metadata
        md_urls: Optional[List[str]] = metadata.get_all("Project-URL", None)
        if md_urls is not None:
            for url in md_urls:
                key, url = url.split(",", maxsplit=1)
                key = key.lower().strip()
                if key in keys:
                    return url.strip()
        self.log_info(f"Unable to find {primary} url in metadata")
        return ""

    def _update_current_version(self, package_info: PackageInfo) -> bool:
        pkg_verson = ""
        release_info = package_info.release_info
        metadata = package_info.metadata
        if release_info is not None:
            self.current_sha = release_info.get("commit_sha", self.current_sha)
            self.git_version = GitVersion(release_info.get("git_version", "?"))
            pkg_verson = release_info.get("package_version", "")
        if "Version" in metadata:
            pkg_verson = metadata["Version"]
        if not pkg_verson:
            self._add_warning("Failed to detect package version")
            return False
        self.current_version = PyVersion(pkg_verson)
        if not self.current_version.is_valid_version():
            self._add_warning("Failed to parse package version")
            return False
        local = self.current_version.local
        if self.current_sha == "?":
            if self.source != PackageSource.GITHUB:
                self.current_sha = "not-specified"
            elif local:
                self.current_sha = local[1:].split(".", 1)[0]
        if not self.git_version.is_valid_version():
            self.git_version = self.current_version.convert_to_git()
        return True

    def _parse_system_dependencies(self, package_info: PackageInfo) -> List[str]:
        rinfo = package_info.release_info
        if rinfo is None:
            return []
        dep_info = rinfo.get("system_dependencies", {})
        for distro_id in DISTRO_ALIASES:
            if distro_id in dep_info:
                if not dep_info[distro_id]:
                    self.log_info(
                        f"Package release_info contains an empty system "
                        f"package definition for linux distro '{distro_id}'"
                    )
                return dep_info[distro_id]
        else:
            self.log_info(
                "Package release_info has no package definition "
                f" for linux distro '{DISTRO_ALIASES[0]}'"
            )
            return []

    async def _update_local_state(self) -> None:
        self.warnings.clear()
        eventloop = self.server.get_event_loop()
        try:
            assert self.virtualenv is not None
            package_info = await eventloop.run_in_thread(
                load_distribution_info, self.virtualenv, self.project_name
            )
        except self.server.error:
            self._add_warning("Failed to parse package info")
        else:
            self.git_version = GitVersion("?")
            self.current_sha = "?"
            self.current_version = PyVersion("?")
            self._detect_update_source(package_info)
            self._update_current_version(package_info)
            self.changelog = self._get_url(["changelog"], package_info)
            self.system_deps = self._parse_system_dependencies(package_info)
        self._is_valid = len(self.warnings) == 0

    async def refresh(self) -> None:
        try:
            if self.source == PackageSource.PIP:
                await self._refresh_pip()
            elif self.source == PackageSource.GITHUB:
                await self._refresh_github()
            else:
                self.log_info("Cannot refresh, package source is unknown")
        except asyncio.CancelledError:
            raise
        except Exception:
            self.log_exc(f"Error Refreshing Python Package: {self.name}")
        self._log_package_info()
        self._save_state()

    async def _refresh_pip(self) -> None:
        # Perform a dry-run install to see if an update is available.
        # Curently this is the most reliable way to fetch the latest
        # version from an index, as we can't assume configurations
        # will use PyPI.
        self.log_info("Requesting package info via PIP...")
        norm_name = normalize_project_name(self.project_name)
        assert self.pip_cmd is not None
        pip_args = f"install -U --quiet --dry-run --no-deps --report - {norm_name}"
        pip_exec = pip_utils.AsyncPipExecutor(self.pip_cmd, self.server)
        resp = await pip_exec.call_pip_with_response(pip_args)
        data: Dict[str, Any] = json_wrapper.loads(resp)
        install_data: List[Dict[str, Any]] = data.get("install", [])
        if not install_data:
            # No update available
            self.upstream_version = self.current_version
            return
        metadata: Dict[str, Any] = install_data[0].get("metadata", {})
        name: str = normalize_project_name(metadata.get("name", ""))
        if len(install_data) > 1 and name != norm_name:
            for inst in install_data[1:]:
                md: Dict[str, Any] = inst.get("metadata", {})
                name = normalize_project_name(md.get("name", ""))
                if name == norm_name:
                    metadata = md
                    break
            else:
                raise self.server.error("Failed to find metadata for package")
        version: str = metadata.get("version", "?")
        self.upstream_version = PyVersion(version)
        if self.current_version < self.upstream_version:
            self.upstream_sha = "update-available"
        else:
            self.upstream_sha = self.current_sha

    async def _refresh_github(self) -> None:
        repo = f"{self.repo_owner}/{self.repo_name}"
        client = self.cmd_helper.get_http_client()
        if self.channel == Channel.DEV:
            resource = f"/repos/{repo}/commits?per_page=1"
            if self.primary_branch is not None:
                resource += f"&sha={self.primary_branch}"
            resp = await client.github_api_request(
                resource, attempts=3, retry_pause_time=.5
            )
            if resp.status_code != 304 and resp.has_error():
                self.log_info(f"Github Request Error - {resp.error}")
                return
            commit_list: List[Dict[str, Any]] = cast(list, resp.json())
            if not commit_list:
                self.log_info("No commits found")
                return
            self.upstream_sha = commit_list[0]["sha"]
            self.upstream_version = self.current_version
            if self.upstream_sha != self.current_sha:
                local_part = f"g{self.upstream_sha[:8]}"
                bumped = self.current_version.bump_local_version(local_part)
                self.upstream_version = bumped
            return
        if self.channel == Channel.STABLE:
            resource = f"repos/{repo}/releases/latest"
        else:
            resource = f"repos/{repo}/releases?per_page=1"
        resp = await client.github_api_request(
            resource, attempts=3, retry_pause_time=.5
        )
        if resp.status_code != 304 and resp.has_error():
            self.log_info(f"Github Request Error - {resp.error}")
            return
        release = resp.json()
        result: Dict[str, Any] = {}
        if isinstance(release, list):
            if release:
                result = release[0]
        else:
            result = release
        if not result:
            self.log_info("No releases found")
            self.upstream_sha = self.current_sha
            self.upstream_version = self.current_version
            return
        self.upstream_version = PyVersion(result["tag_name"])
        if self.upstream_version > self.current_version:
            self.upstream_sha = "update-available"

    async def update(self, rollback: bool = False) -> bool:
        project_name = normalize_project_name(self.project_name)
        assert self.pip_cmd is not None
        pip_args: str
        if not self.upstream_version.is_valid_version():
            # Can't update without a valid upstream
            return False
        pip_exec = pip_utils.AsyncPipExecutor(
            self.pip_cmd, self.server, self.cmd_helper.notify_update_response
        )
        current_ref = self.current_version.tag
        if self.source == PackageSource.PIP:
            # We can't depend on the SHA being available for PyPI packages,
            # so we must compare versions
            if (
                self.current_version.is_valid_version() and
                self.upstream_version <= self.current_version
            ):
                return False
            pip_args = f"install -U {project_name}"
            if rollback:
                pip_args += f"=={self.rollback_ref}"
        elif self.source == PackageSource.GITHUB:
            if self.current_sha == self.upstream_sha:
                return False
            repo = f"{self.repo_owner}/{self.repo_name}"
            pip_args = f"install -U git+https://github.com/{repo}"
            if rollback:
                pip_args += f"@{self.rollback_ref}"
            elif self.channel == "dev":
                current_ref = self.current_sha
                if self.primary_branch is not None:
                    pip_args += f"@{self.primary_branch}"
            else:
                pip_args += f"@{self.upstream_version.tag}"
        else:
            raise self.server.error("Cannot update, package source is unknown")
        await self._update_pip(pip_exec)
        sys_deps = self.system_deps
        source = self.source.name
        self.notify_status(f"Updating Python Package {self.name} from {source}...")
        await pip_exec.call_pip(pip_args, 3600, sys_env_vars=self.pip_env_vars)
        await self._update_local_state()
        if not rollback:
            self.rollback_ref = current_ref
        self.upstream_sha = self.current_sha
        self.upstream_version = self.current_version
        await self._update_sys_deps(sys_deps)
        self._log_package_info()
        self._save_state()
        await self.restart_service()
        self.notify_status("Update Finished...", is_complete=True)
        return True

    async def recover(
        self, hard: bool = False, force_dep_update: bool = False
    ) -> None:
        pass

    async def rollback(self) -> bool:
        if self.rollback_ref == "?":
            return False
        await self.update(rollback=True)
        return True

    async def _update_sys_deps(self, prev_deps: List[str]) -> None:
        new_deps = self.system_deps
        deps_diff = list(set(new_deps) - set(prev_deps))
        if deps_diff:
            await self._install_packages(deps_diff)

    def _log_package_info(self) -> None:
        logging.info(
            f"Python Package {self.name} detected:\n"
            f"Channel: {self.channel}\n"
            f"Package Source: {self.source.name}\n"
            f"Repo Owner: {self.repo_owner}\n"
            f"Repo Name: {self.repo_name}\n"
            f"Repo URL: {self.repo_url}\n"
            f"Changelog URL: {self.changelog}\n"
            f"Full Version String: {self.current_version.full_version}\n"
            f"Current Version: {self.current_version.short_version}\n"
            f"Current Commit SHA: {self.current_sha}\n"
            f"Upstream Version: {self.upstream_version.short_version}\n"
            f"Upstream Commit SHA: {self.upstream_sha}\n"
            f"Converted Git Version: {self.git_version}\n"
            f"Rollback Ref: {self.rollback_ref}\n"
        )



# Zeroconf registration implementation for Moonraker
#
# Copyright (C) 2021  Clifford Roche <clifford.roche@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license.
from __future__ import annotations
import socket
import asyncio
import logging
import ipaddress
import random
import uuid
from itertools import cycle
from email.utils import formatdate
from zeroconf import IPVersion
from zeroconf.asyncio import AsyncServiceInfo, AsyncZeroconf
from ..common import RequestType, TransportType

from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterator,
    List,
    Optional,
    Tuple
)

if TYPE_CHECKING:
    from ..confighelper import ConfigHelper
    from ..common import WebRequest
    from .application import MoonrakerApp
    from .machine import Machine

ZC_SERVICE_TYPE = "_moonraker._tcp.local."

class AsyncRunner:
    def __init__(self, ip_version: IPVersion) -> None:
        self.ip_version = ip_version
        self.aiozc: Optional[AsyncZeroconf] = None

    async def register_services(self, infos: List[AsyncServiceInfo]) -> None:
        self.aiozc = AsyncZeroconf(ip_version=self.ip_version)
        tasks = [
            self.aiozc.async_register_service(info, allow_name_change=True)
            for info in infos
        ]
        background_tasks = await asyncio.gather(*tasks)
        await asyncio.gather(*background_tasks)

    async def unregister_services(self, infos: List[AsyncServiceInfo]) -> None:
        assert self.aiozc is not None
        tasks = [self.aiozc.async_unregister_service(info) for info in infos]
        background_tasks = await asyncio.gather(*tasks)
        await asyncio.gather(*background_tasks)
        await self.aiozc.async_close()

    async def update_services(self, infos: List[AsyncServiceInfo]) -> None:
        assert self.aiozc is not None
        tasks = [self.aiozc.async_update_service(info) for info in infos]
        background_tasks = await asyncio.gather(*tasks)
        await asyncio.gather(*background_tasks)


class ZeroconfRegistrar:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        hi = self.server.get_host_info()
        self.mdns_name = config.get("mdns_hostname", hi["hostname"])
        addr: str = hi["address"]
        self.ip_version = IPVersion.All
        if addr.lower() == "all":
            addr = "::"
        else:
            addr_obj = ipaddress.ip_address(addr)
            self.ip_version = (
                IPVersion.V4Only if addr_obj.version == 4 else IPVersion.V6Only
            )
        self.runner = AsyncRunner(self.ip_version)
        self.cfg_addr = addr
        self.bound_all = addr in ["0.0.0.0", "::"]
        if self.bound_all:
            self.server.register_event_handler(
                "machine:net_state_changed", self._update_service)
        self.ssdp_server: Optional[SSDPServer] = None
        if config.getboolean("enable_ssdp", False):
            self.ssdp_server = SSDPServer(config)

    async def component_init(self) -> None:
        logging.info("Starting Zeroconf services")
        app: MoonrakerApp = self.server.lookup_component("application")
        machine: Machine = self.server.lookup_component("machine")
        app_args = self.server.get_app_args()
        instance_uuid: str = app_args["instance_uuid"]
        if (
            machine.get_provider_type().startswith("systemd") and
            "unit_name" in machine.get_moonraker_service_info()
        ):
            # Use the name of the systemd service unit to identify service
            instance_name = machine.unit_name.capitalize()
        else:
            # Use the UUID.  First 8 hex digits should be unique enough
            instance_name = f"Moonraker-{instance_uuid[:8]}"
        hi = self.server.get_host_info()
        host = self.mdns_name
        zc_service_props = {
            "uuid": instance_uuid,
            "https_port": hi["ssl_port"] if app.https_enabled() else "",
            "version": app_args["software_version"],
            "route_prefix": app.route_prefix
        }
        if self.bound_all:
            if not host:
                host = machine.public_ip
            network = machine.get_system_info()["network"]
            addresses: List[bytes] = [x for x in self._extract_ip_addresses(network)]
        else:
            if not host:
                host = self.cfg_addr
            host_addr = ipaddress.ip_address(self.cfg_addr)
            addresses = [host_addr.packed]
        zc_service_name = f"{instance_name} @ {host}.{ZC_SERVICE_TYPE}"
        server_name = self.mdns_name or instance_name.lower()
        self.service_info = AsyncServiceInfo(
            ZC_SERVICE_TYPE,
            zc_service_name,
            addresses=addresses,
            port=hi["port"],
            properties=zc_service_props,
            server=f"{server_name}.local.",
        )
        await self.runner.register_services([self.service_info])
        if self.ssdp_server is not None:
            addr = self.cfg_addr if not self.bound_all else machine.public_ip
            if not addr:
                addr = f"{self.mdns_name}.local"
            name = f"{instance_name} ({host})"
            if len(name) > 64:
                name = instance_name
            await self.ssdp_server.start()
            self.ssdp_server.register_service(name, addr, hi["port"])

    async def close(self) -> None:
        await self.runner.unregister_services([self.service_info])
        if self.ssdp_server is not None:
            await self.ssdp_server.stop()

    async def _update_service(self, network: Dict[str, Any]) -> None:
        if self.bound_all:
            addresses = [x for x in self._extract_ip_addresses(network)]
            self.service_info.addresses = addresses
            await self.runner.update_services([self.service_info])

    def _extract_ip_addresses(self, network: Dict[str, Any]) -> Iterator[bytes]:
        for ifname, ifinfo in network.items():
            for addr_info in ifinfo["ip_addresses"]:
                if addr_info["is_link_local"]:
                    continue
                addr_obj = ipaddress.ip_address(addr_info["address"])
                ver = addr_obj.version
                if (
                    (self.ip_version == IPVersion.V4Only and ver == 6) or
                    (self.ip_version == IPVersion.V6Only and ver == 4)
                ):
                    continue
                yield addr_obj.packed


SSDP_ADDR = ("239.255.255.250", 1900)
SSDP_SERVER_ID = "Moonraker SSDP/UPNP Server"
SSDP_MAX_AGE = 1800
SSDP_DEVICE_TYPE = "urn:arksine.github.io:device:Moonraker:1"
SSDP_DEVICE_XML = """
<?xml version="1.0"?>
<root xmlns="urn:schemas-upnp-org:device-1-0" configId="{config_id}">
    <specVersion>
        <major>2</major>
        <minor>0</minor>
    </specVersion>
    <device>
        <deviceType>{device_type}</deviceType>
        <friendlyName>{friendly_name}</friendlyName>
        <manufacturer>Arksine</manufacturer>
        <manufacturerURL>https://github.com/Arksine/moonraker</manufacturerURL>
        <modelDescription>API Server for Klipper</modelDescription>
        <modelName>Moonraker</modelName>
        <modelNumber>{model_number}</modelNumber>
        <modelURL>https://github.com/Arksine/moonraker</modelURL>
        <serialNumber>{serial_number}</serialNumber>
        <UDN>uuid:{device_uuid}</UDN>
        <presentationURL>{presentation_url}</presentationURL>
    </device>
</root>
""".strip()

class SSDPServer(asyncio.protocols.DatagramProtocol):
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.unique_id = uuid.UUID(self.server.get_app_args()["instance_uuid"])
        self.name: str = "Moonraker"
        self.base_url: str = ""
        self.response_headers: List[str] = []
        self.registered: bool = False
        self.running: bool = False
        self.close_fut: Optional[asyncio.Future] = None
        self.response_handle: Optional[asyncio.TimerHandle] = None
        eventloop = self.server.get_event_loop()
        self.boot_id = int(eventloop.get_loop_time())
        self.config_id = 1
        self.ad_timer = eventloop.register_timer(self._advertise_presence)
        self.server.register_endpoint(
            "/server/zeroconf/ssdp",
            RequestType.GET,
            self._handle_xml_request,
            transports=TransportType.HTTP,
            wrap_result=False,
            content_type="application/xml",
            auth_required=False
        )

    def _create_ssdp_socket(
        self,
        source_addr: Tuple[str, int] = ("0.0.0.0", 0),
        target_addr: Tuple[str, int] = SSDP_ADDR
    ) -> socket.socket:
        sock = socket.socket(socket.AF_INET, socket.SOCK_DGRAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        try:
            sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEPORT, 1)
        except AttributeError:
            pass
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_BROADCAST, 1)
        source_ip = socket.inet_aton(source_addr[0])
        target_ip = socket.inet_aton(target_addr[0])
        ip_combo = target_ip + source_ip
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_IF, source_ip)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_MULTICAST_TTL, 2)
        sock.setsockopt(socket.IPPROTO_IP, socket.IP_ADD_MEMBERSHIP, ip_combo)
        return sock

    async def start(self) -> None:
        if self.running:
            return
        try:
            sock = self._create_ssdp_socket()
            sock.settimeout(0)
            sock.setblocking(False)
            sock.bind(("", SSDP_ADDR[1]))
            _loop = asyncio.get_running_loop()
            ret = await _loop.create_datagram_endpoint(lambda: self, sock=sock)
            self.transport, _ = ret
        except (socket.error, OSError):
            return
        self.running = True

    async def stop(self) -> None:
        if not self.running:
            return
        self.running = False
        self.ad_timer.stop()
        if self.response_handle is not None:
            self.response_handle.cancel()
            self.response_handle = None
        if self.transport.is_closing():
            logging.info("Transport already closing")
            return
        for notification in self._build_notifications("ssdp:byebye"):
            self.transport.sendto(notification, SSDP_ADDR)
        self.close_fut = self.server.get_event_loop().create_future()
        self.transport.close()
        try:
            await asyncio.wait_for(self.close_fut, 2.)
        except asyncio.TimeoutError:
            pass
        self.close_fut = None

    def register_service(
        self, name: str, host_name_or_ip: str, port: int
    ) -> None:
        if len(name) > 64:
            name = name[:64]
        self.name = name
        app: MoonrakerApp = self.server.lookup_component("application")
        self.base_url = f"http://{host_name_or_ip}:{port}{app.route_prefix}"
        self.response_headers = [
            f"USN: uuid:{self.unique_id}::upnp:rootdevice",
            f"LOCATION: {self.base_url}/server/zeroconf/ssdp",
            "ST: upnp:rootdevice",
            "EXT:",
            f"SERVER: {SSDP_SERVER_ID}",
            f"CACHE-CONTROL: max-age={SSDP_MAX_AGE}",
            f"BOOTID.UPNP.ORG: {self.boot_id}",
            f"CONFIGID.UPNP.ORG: {self.config_id}",
        ]
        self.registered = True
        advertisements = self._build_notifications("ssdp:alive")
        if self.running:
            for ad in advertisements:
                self.transport.sendto(ad, SSDP_ADDR)
        self.advertisements = cycle(advertisements)
        self.ad_timer.start()

    async def _handle_xml_request(self, web_request: WebRequest) -> str:
        if not self.registered:
            raise self.server.error("Moonraker SSDP Device not registered", 404)
        app_args = self.server.get_app_args()
        return SSDP_DEVICE_XML.format(
            device_type=SSDP_DEVICE_TYPE,
            config_id=str(self.config_id),
            friendly_name=self.name,
            model_number=app_args["software_version"],
            serial_number=self.unique_id.hex,
            device_uuid=str(self.unique_id),
            presentation_url=self.base_url
        )

    def _advertise_presence(self, eventtime: float) -> float:
        if self.running and self.registered:
            cur_ad = next(self.advertisements)
            self.transport.sendto(cur_ad, SSDP_ADDR)
        delay = random.uniform(SSDP_MAX_AGE / 6., SSDP_MAX_AGE / 3.)
        return eventtime + delay

    def connection_made(
        self, transport: asyncio.transports.BaseTransport
    ) -> None:
        logging.debug("SSDP Server Connected")

    def connection_lost(self, exc: Exception | None) -> None:
        logging.debug("SSDP Server Disconnected")
        if self.close_fut is not None:
            self.close_fut.set_result(None)

    def pause_writing(self) -> None:
        logging.debug("SSDP Pause Writing Requested")

    def resume_writing(self) -> None:
        logging.debug("SSDP Resume Writing Requested")

    def datagram_received(self, data: bytes, addr: tuple[str | Any, int]) -> None:
        if not self.registered:
            return
        try:
            parts = data.decode().split("\r\n\r\n", maxsplit=1)
            header = parts[0]
        except ValueError:
            logging.exception("Data Decode Error")
            return
        hlines = header.splitlines()
        ssdp_command = hlines[0].strip()
        headers = {}
        for line in hlines[1:]:
            parts = line.strip().split(":", maxsplit=1)
            if len(parts) < 2:
                continue
            headers[parts[0].upper()] = parts[1].strip()
        if (
            ssdp_command != "M-SEARCH * HTTP/1.1" or
            headers.get("MAN") != '"ssdp:discover"'
        ):
            # Not a discovery request
            return
        if headers.get("ST") not in ["upnp:rootdevice", "ssdp:all"]:
            # Service Type doesn't apply
            return
        if self.response_handle is not None:
            # response in progress
            return
        if "MX" in headers:
            delay_time = random.uniform(0, float(headers["MX"]))
            eventloop = self.server.get_event_loop()
            self.response_handle = eventloop.delay_callback(
                delay_time, self._respond_to_discovery, addr
            )
        else:
            self._respond_to_discovery(addr)

    def _respond_to_discovery(self, addr: tuple[str | Any, int]) -> None:
        if not self.running:
            return
        self.response_handle = None
        response: List[str] = ["HTTP/1.1 200 OK"]
        response.extend(self.response_headers)
        response.append(f"DATE: {formatdate(usegmt=True)}")
        response.extend(["", ""])
        self.transport.sendto("\r\n".join(response).encode(), addr)

    def _build_notifications(self, nts: str) -> List[bytes]:
        notifications: List[bytes] = []
        notify_types = [
            ("upnp:rootdevice", f"uuid:{self.unique_id}::upnp:rootdevice"),
            (f"uuid:{self.unique_id}", f"uuid:{self.unique_id}"),
            (SSDP_DEVICE_TYPE, f"uuid:{self.unique_id}::{SSDP_DEVICE_TYPE}")
        ]
        for (nt, usn) in notify_types:
            notifications.append(
                "\r\n".join([
                    "NOTIFY * HTTP/1.1",
                    f"HOST: {SSDP_ADDR[0]}:{SSDP_ADDR[1]}",
                    f"NTS: {nts}",
                    f"NT: {nt}",
                    f"USN: {usn}",
                    f"LOCATION: {self.base_url}/server/zeroconf/ssdp",
                    "EXT:",
                    f"SERVER: {SSDP_SERVER_ID}",
                    f"CACHE-CONTROL: max-age={SSDP_MAX_AGE}",
                    f"BOOTID.UPNP.ORG: {self.boot_id}",
                    f"CONFIGID.UPNP.ORG: {self.config_id}",
                    "",
                    ""
                ]).encode()
            )
        return notifications

    def error_received(self, exc: Exception) -> None:
        logging.info(f"SSDP Server Error: {exc}")


def load_component(config: ConfigHelper) -> ZeroconfRegistrar:
    return ZeroconfRegistrar(config)



# Wrapper around the asyncio eventloop
#
# Copyright (C) 2021 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import os
import contextlib
import asyncio
import inspect
import functools
import socket
import time
import logging
from typing import (
    TYPE_CHECKING,
    Awaitable,
    Callable,
    Optional,
    Tuple,
    TypeVar,
    Union
)

_uvl_var = os.getenv("MOONRAKER_ENABLE_UVLOOP", "y").lower()
_uvl_enabled = False
if _uvl_var in ["y", "yes", "true"]:
    with contextlib.suppress(ImportError):
        import uvloop  # type: ignore
        asyncio.set_event_loop_policy(uvloop.EventLoopPolicy())
        _uvl_enabled = True

if TYPE_CHECKING:
    from asyncio import AbstractEventLoop
    _T = TypeVar("_T")
    FlexCallback = Callable[..., Optional[Awaitable]]
    TimerCallback = Callable[[float], Union[float, Awaitable[float]]]

class EventLoop:
    UVLOOP_ENABLED = _uvl_enabled
    TimeoutError = asyncio.TimeoutError
    def __init__(self) -> None:
        self.reset()

    @property
    def asyncio_loop(self) -> AbstractEventLoop:
        return self.aioloop

    def reset(self) -> None:
        self.aioloop = asyncio.get_running_loop()
        self.add_signal_handler = self.aioloop.add_signal_handler
        self.remove_signal_handler = self.aioloop.remove_signal_handler
        self.add_reader = self.aioloop.add_reader
        self.add_writer = self.aioloop.add_writer
        self.remove_reader = self.aioloop.remove_reader
        self.remove_writer = self.aioloop.remove_writer
        self.get_loop_time = self.aioloop.time
        self.create_future = self.aioloop.create_future
        self.create_task = self.aioloop.create_task
        self.call_at = self.aioloop.call_at
        self.set_debug = self.aioloop.set_debug
        self.is_running = self.aioloop.is_running

    def _create_new_loop(self) -> asyncio.AbstractEventLoop:
        for _ in range(5):
            # Sometimes the new loop does not properly instantiate.
            # Give 5 attempts before raising an exception
            new_loop = asyncio.new_event_loop()
            if not new_loop.is_closed():
                break
            logging.info("Failed to create open eventloop, "
                         "retyring in .5 seconds...")
            time.sleep(.5)
        else:
            raise RuntimeError("Unable to create new open eventloop")
        asyncio.set_event_loop(new_loop)
        return new_loop

    def register_callback(self,
                          callback: FlexCallback,
                          *args,
                          **kwargs
                          ) -> None:
        async def _wrapper():
            try:
                ret = callback(*args, **kwargs)
                if inspect.isawaitable(ret):
                    await ret
            except asyncio.CancelledError:
                raise
            except Exception:
                logging.exception("Error Running Callback")
        self.aioloop.create_task(_wrapper())

    def delay_callback(self,
                       delay: float,
                       callback: FlexCallback,
                       *args,
                       **kwargs
                       ) -> asyncio.TimerHandle:
        return self.aioloop.call_later(
            delay, self.register_callback,
            functools.partial(callback, *args, **kwargs)
        )

    def register_timer(self, callback: TimerCallback):
        return FlexTimer(self, callback)

    def run_in_thread(self,
                      callback: Callable[..., _T],
                      *args
                      ) -> Awaitable[_T]:
        return self.aioloop.run_in_executor(None, callback, *args)

    async def create_socket_connection(
        self, address: Tuple[str, int], timeout: Optional[float] = None
    ) -> socket.socket:
        host, port = address
        """
        async port of socket.create_connection()
        """
        loop = self.aioloop
        err = None
        ainfo = await loop.getaddrinfo(
            host, port, family=0, type=socket.SOCK_STREAM
        )
        for res in ainfo:
            af, socktype, proto, _cannon_name, _sa = res
            sock = None
            try:
                sock = socket.socket(af, socktype, proto)
                sock.settimeout(0)
                sock.setblocking(False)
                await asyncio.wait_for(
                    loop.sock_connect(sock, (host, port)), timeout
                )
                # Break explicitly a reference cycle
                err = None
                return sock
            except (socket.error, asyncio.TimeoutError) as _:
                err = _
                if sock is not None:
                    loop.remove_writer(sock.fileno())
                    sock.close()
        if err is not None:
            try:
                raise err
            finally:
                # Break explicitly a reference cycle
                err = None
        else:
            raise socket.error("getaddrinfo returns an empty list")

    def close(self):
        self.aioloop.close()

class FlexTimer:
    def __init__(self,
                 eventloop: EventLoop,
                 callback: TimerCallback
                 ) -> None:
        self.eventloop = eventloop
        self.callback = callback
        self.timer_handle: Optional[asyncio.TimerHandle] = None
        self.timer_task: Optional[asyncio.Task] = None
        self.running: bool = False

    def in_callback(self) -> bool:
        return self.timer_task is not None and not self.timer_task.done()

    def start(self, delay: float = 0.):
        if self.running:
            return
        self.running = True
        if self.in_callback():
            return
        call_time = self.eventloop.get_loop_time() + delay
        self.timer_handle = self.eventloop.call_at(
            call_time, self._schedule_task)

    def stop(self):
        if not self.running:
            return
        self.running = False
        if self.timer_handle is not None:
            self.timer_handle.cancel()
            self.timer_handle = None

    async def wait_timer_done(self) -> None:
        if self.timer_task is None:
            return
        await self.timer_task

    def _schedule_task(self):
        self.timer_handle = None
        self.timer_task = self.eventloop.create_task(self._call_wrapper())

    def is_running(self) -> bool:
        return self.running

    async def _call_wrapper(self):
        if not self.running:
            return
        try:
            ret = self.callback(self.eventloop.get_loop_time())
            if isinstance(ret, Awaitable):
                ret = await ret
        except Exception:
            self.running = False
            raise
        finally:
            self.timer_task = None
        if self.running:
            self.timer_handle = self.eventloop.call_at(ret, self._schedule_task)



# Methods to create IOCTL requests
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import ctypes
from typing import Union, Type, TYPE_CHECKING

"""
This module contains of Python port of the macros avaialble in
"/include/uapi/asm-generic/ioctl.h" from the linux kernel.
"""

if TYPE_CHECKING:
    IOCParamSize = Union[int, str, Type[ctypes._CData]]

_IOC_NRBITS = 8
_IOC_TYPEBITS = 8

# NOTE: The following could be platform specific.
_IOC_SIZEBITS = 14
_IOC_DIRBITS = 2

_IOC_NRMASK = (1 << _IOC_NRBITS) - 1
_IOC_TYPEMASK = (1 << _IOC_TYPEBITS) - 1
_IOC_SIZEMASK = (1 << _IOC_SIZEBITS) - 1
_IOC_DIRMASK = (1 << _IOC_DIRBITS) - 1

_IOC_NRSHIFT = 0
_IOC_TYPESHIFT = _IOC_NRSHIFT + _IOC_NRBITS
_IOC_SIZESHIFT = _IOC_TYPESHIFT + _IOC_TYPEBITS
_IOC_DIRSHIFT = _IOC_SIZESHIFT + _IOC_SIZEBITS

# The constants below may also be platform specific
IOC_NONE = 0
IOC_WRITE = 1
IOC_READ = 2

def _check_value(val: int, name: str, maximum: int):
    if val > maximum:
        raise ValueError(f"Value '{val}' for '{name}' exceeds max of {maximum}")

def _IOC_TYPECHECK(param_size: IOCParamSize) -> int:
    if isinstance(param_size, int):
        return param_size
    elif isinstance(param_size, bytearray):
        return len(param_size)
    elif isinstance(param_size, str):
        ctcls = getattr(ctypes, param_size)
        return ctypes.sizeof(ctcls)
    return ctypes.sizeof(param_size)

def IOC(direction: int, cmd_type: int, cmd_number: int, param_size: int) -> int:
    _check_value(direction, "direction", _IOC_DIRMASK)
    _check_value(cmd_type, "cmd_type", _IOC_TYPEMASK)
    _check_value(cmd_number, "cmd_number", _IOC_NRMASK)
    _check_value(param_size, "ioc_size", _IOC_SIZEMASK)
    return (
        (direction << _IOC_DIRSHIFT) |
        (param_size << _IOC_SIZESHIFT) |
        (cmd_type << _IOC_TYPESHIFT) |
        (cmd_number << _IOC_NRSHIFT)
    )

def IO(cmd_type: int, cmd_number: int) -> int:
    return IOC(IOC_NONE, cmd_type, cmd_number, 0)

def IOR(cmd_type: int, cmd_number: int, param_size: IOCParamSize) -> int:
    return IOC(IOC_READ, cmd_type, cmd_number, _IOC_TYPECHECK(param_size))

def IOW(cmd_type: int, cmd_number: int, param_size: IOCParamSize) -> int:
    return IOC(IOC_WRITE, cmd_type, cmd_number, _IOC_TYPECHECK(param_size))

def IOWR(cmd_type: int, cmd_number: int, param_size: IOCParamSize) -> int:
    return IOC(IOC_READ | IOC_WRITE, cmd_type, cmd_number, _IOC_TYPECHECK(param_size))



# Wheel Setup Script for generating metadata
#
# Copyright (C) 2023 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import pathlib
import subprocess
import shlex
import json
import shutil
from datetime import datetime, timezone
from typing import Dict, Any, TYPE_CHECKING

if TYPE_CHECKING:
    from pdm.backend.hooks.base import Context

__package_name__ = "moonraker"
__dependencies__ = "scripts/system-dependencies.json"

def _run_git_command(cmd: str) -> str:
    prog = shlex.split(cmd)
    process = subprocess.Popen(
        prog, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    ret, err = process.communicate()
    retcode = process.wait()
    if retcode == 0:
        return ret.strip().decode()
    return ""

def get_commit_sha(source_path: pathlib.Path) -> str:
    cmd = f"git -C {source_path} rev-parse HEAD"
    return _run_git_command(cmd)

def retrieve_git_version(source_path: pathlib.Path) -> str:
    cmd = f"git -C {source_path} describe --always --tags --long --dirty"
    return _run_git_command(cmd)

def pdm_build_initialize(context: Context) -> None:
    context.ensure_build_dir()
    proj_name: str = context.config.metadata['name']
    build_dir = pathlib.Path(context.build_dir)
    pkg_path = build_dir.joinpath(__package_name__)
    pkg_path.mkdir(parents=True, exist_ok=True)
    rinfo_path: pathlib.Path = pkg_path.joinpath("release_info")
    rinfo_data: str = ""
    if context.root.joinpath(".git").exists():
        build_ver: str = context.config.metadata['version']
        build_time = datetime.now(timezone.utc)
        urls: Dict[str, str] = context.config.metadata['urls']
        release_info: Dict[str, Any] = {
            "project_name": proj_name,
            "package_name": __package_name__,
            "urls": {key.lower(): val for key, val in urls.items()},
            "package_version": build_ver,
            "git_version": retrieve_git_version(context.root),
            "commit_sha": get_commit_sha(context.root),
            "build_time": datetime.isoformat(build_time, timespec="seconds")
        }
        if __dependencies__:
            deps = pathlib.Path(context.root).joinpath(__dependencies__)
            if deps.is_file():
                dep_info: Dict[str, Any] = json.loads(deps.read_bytes())
                release_info["system_dependencies"] = dep_info
        # Write the release info to both the package and the data path
        rinfo_data = json.dumps(release_info, indent=4)
        rinfo_path.write_text(rinfo_data)
        scripts_path: pathlib.Path = context.root.joinpath("scripts")
        scripts_dest: pathlib.Path = pkg_path.joinpath("scripts")
        scripts_dest.mkdir()
        for item in scripts_path.iterdir():
            if item.name in ("__pycache__", "python_wheels"):
                continue
            if item.is_dir():
                shutil.copytree(str(item), str(scripts_dest.joinpath(item.name)))
            else:
                shutil.copy2(str(item), str(scripts_dest))
    git_ignore = build_dir.joinpath(".gitignore")
    if git_ignore.is_file():
        git_ignore.unlink()



# Wrapper around Tornado's HTTP Client with a "requests-like" interface
#
# Copyright (C) 2022 Eric Callahan <arksine.code@gmail.com>
#
# This file may be distributed under the terms of the GNU GPLv3 license

from __future__ import annotations
import re
import time
import asyncio
import pathlib
import tempfile
import logging
import copy
from ..utils import ServerError
from ..utils import json_wrapper as jsonw
from tornado.escape import url_unescape
from tornado.httpclient import AsyncHTTPClient, HTTPRequest, HTTPError
from tornado.httputil import HTTPHeaders
from typing import (
    TYPE_CHECKING,
    Callable,
    Optional,
    Tuple,
    Union,
    Dict,
    List,
    Any
)
if TYPE_CHECKING:
    from ..server import Server
    from ..confighelper import ConfigHelper
    from io import BufferedWriter
    StrOrPath = Union[str, pathlib.Path]

MAX_BODY_SIZE = 512 * 1024 * 1024
AsyncHTTPClient.configure(
    None, defaults=dict(user_agent="Moonraker"),
    max_body_size=MAX_BODY_SIZE
)

GITHUB_PREFIX = "https://api.github.com/"

class HttpClient:
    def __init__(self, config: ConfigHelper) -> None:
        self.server = config.get_server()
        self.client = AsyncHTTPClient()
        self.response_cache: Dict[str, HttpResponse] = {}

        self.gh_rate_limit: Optional[int] = None
        self.gh_limit_remaining: Optional[int] = None
        self.gh_limit_reset_time: Optional[float] = None

    def register_cached_url(
        self,
        url: str,
        etag: Optional[str] = None,
        last_modified: Optional[str] = None
    ) -> None:
        headers = HTTPHeaders()
        if etag is not None:
            headers["etag"] = etag
        if last_modified is not None:
            headers["last-modified"] = last_modified
        if len(headers) == 0:
            raise self.server.error(
                "Either an Etag or Last Modified Date must be specified")
        empty_resp = HttpResponse(url, url, 200, b"", headers, None)
        self.response_cache[url] = empty_resp

    async def request(
        self,
        method: str,
        url: str,
        body: Optional[Union[bytes, str, List[Any], Dict[str, Any]]] = None,
        headers: Optional[Dict[str, Any]] = None,
        connect_timeout: float = 5.,
        request_timeout: float = 10.,
        attempts: int = 1,
        retry_pause_time: float = .1,
        enable_cache: bool = False,
        send_etag: bool = True,
        send_if_modified_since: bool = True,
        basic_auth_user: Optional[str] = None,
        basic_auth_pass: Optional[str] = None
    ) -> HttpResponse:
        cache_key = url.split("?", 1)[0]
        method = method.upper()
        # prepare the body if required
        req_headers: Dict[str, Any] = {}
        if isinstance(body, (list, dict)):
            body = jsonw.dumps(body)
            req_headers["Content-Type"] = "application/json"
        cached: Optional[HttpResponse] = None
        if enable_cache:
            cached = self.response_cache.get(cache_key)
            if cached is not None and send_etag:
                if cached.etag is not None and send_etag:
                    req_headers["If-None-Match"] = cached.etag
                if cached.last_modified and send_if_modified_since:
                    req_headers["If-Modified-Since"] = cached.last_modified
        if headers is not None:
            headers.update(req_headers)
        elif req_headers:
            headers = req_headers

        timeout = 1 + connect_timeout + request_timeout
        req_args: Dict[str, Any] = dict(
            body=body,
            request_timeout=request_timeout,
            connect_timeout=connect_timeout
        )
        if basic_auth_user is not None:
            assert basic_auth_pass is not None
            req_args["auth_username"] = basic_auth_user
            req_args["auth_password"] = basic_auth_pass
            req_args["auth_mode"] = "basic"
        request = HTTPRequest(url, method, headers, **req_args)
        err: Optional[BaseException] = None
        for i in range(attempts):
            if i:
                await asyncio.sleep(retry_pause_time)
            try:
                fut = self.client.fetch(request, raise_error=False)
                resp = await asyncio.wait_for(fut, timeout)
            except asyncio.CancelledError:
                raise
            except Exception as e:
                err = e
            else:
                err = resp.error
                if resp.code == 304:
                    err = None
                    if cached is None:
                        if enable_cache:
                            logging.info(
                                "Request returned 304, however no cached "
                                "item was found")
                        result = b""
                    else:
                        logging.debug(f"Request returned from cache: {url}")
                        result = cached.content
                elif resp.error is not None and attempts - i != 1:
                    continue
                else:
                    result = resp.body
                ret = HttpResponse(
                    url, resp.effective_url, resp.code, result,
                    resp.headers, err
                )
                break
        else:
            ret = HttpResponse(url, url, 500, b"", HTTPHeaders(), err)
        if enable_cache and ret.is_cachable():
            logging.debug(f"Caching HTTP Response: {url}")
            self.response_cache[cache_key] = ret
        else:
            self.response_cache.pop(cache_key, None)
        return ret

    async def get(
        self, url: str, headers: Optional[Dict[str, Any]] = None, **kwargs
    ) -> HttpResponse:
        if "enable_cache" not in kwargs:
            kwargs["enable_cache"] = True
        return await self.request("GET", url, None, headers, **kwargs)

    async def post(
        self,
        url: str,
        body: Union[str, List[Any], Dict[str, Any]] = "",
        headers: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> HttpResponse:
        return await self.request("POST", url, body, headers, **kwargs)

    async def delete(
        self,
        url: str,
        headers: Optional[Dict[str, Any]] = None,
        **kwargs
    ) -> HttpResponse:
        return await self.request("DELETE", url, None, headers, **kwargs)

    async def github_api_request(
        self,
        resource: str,
        attempts: int = 1,
        retry_pause_time: float = .1
    ) -> HttpResponse:
        url = f"{GITHUB_PREFIX}{resource.strip('/')}"
        if (
            self.gh_limit_reset_time is not None and
            self.gh_limit_remaining == 0
        ):
            curtime = time.time()
            if curtime < self.gh_limit_reset_time:
                reset_time = time.ctime(self.gh_limit_reset_time)
                raise self.server.error(
                    f"GitHub Rate Limit Reached\n"
                    f"Request: {url}\n"
                    f"Limit Reset Time: {reset_time}"
                )
        headers = {"Accept": "application/vnd.github.v3+json"}
        resp = await self.get(
            url, headers, attempts=attempts,
            retry_pause_time=retry_pause_time)
        resp_hdrs = resp.headers
        if 'X-Ratelimit-Limit' in resp_hdrs:
            self.gh_rate_limit = int(resp_hdrs['X-Ratelimit-Limit'])
            self.gh_limit_remaining = int(
                resp_hdrs['X-Ratelimit-Remaining'])
            self.gh_limit_reset_time = float(
                resp_hdrs['X-Ratelimit-Reset'])
        return resp

    def github_api_stats(self) -> Dict[str, Any]:
        return {
            'github_rate_limit': self.gh_rate_limit,
            'github_requests_remaining': self.gh_limit_remaining,
            'github_limit_reset_time': self.gh_limit_reset_time,
        }

    async def get_file(
        self,
        url: str,
        content_type: str,
        connect_timeout: float = 5.,
        request_timeout: float = 180.,
        attempts: int = 1,
        retry_pause_time: float = .1,
        enable_cache: bool = False,
    ) -> bytes:
        headers = {"Accept": content_type}
        resp = await self.get(
            url, headers, connect_timeout=connect_timeout,
            request_timeout=request_timeout, attempts=attempts,
            retry_pause_time=retry_pause_time, enable_cache=enable_cache
        )
        resp.raise_for_status()
        return resp.content

    async def download_file(
        self,
        url: str,
        content_type: str,
        destination_path: Optional[StrOrPath] = None,
        download_size: int = -1,
        progress_callback: Optional[Callable[[int, int, int], None]] = None,
        connect_timeout: float = 5.,
        request_timeout: float = 180.,
        attempts: int = 1,
        retry_pause_time: float = 1.
    ) -> pathlib.Path:
        for i in range(attempts):
            dl = StreamingDownload(
                self.server, destination_path, download_size,
                progress_callback)
            try:
                fut = self.client.fetch(
                    url, headers={"Accept": content_type},
                    connect_timeout=connect_timeout,
                    request_timeout=request_timeout,
                    streaming_callback=dl.on_chunk_recd,
                    header_callback=dl.on_headers_recd)
                timeout = connect_timeout + request_timeout + 1.
                resp = await asyncio.wait_for(fut, timeout)
            except asyncio.CancelledError:
                raise
            except Exception:
                if i + 1 == attempts:
                    raise
                await asyncio.sleep(retry_pause_time)
                continue
            finally:
                await dl.close()
            if resp.code < 400:
                return dl.dest_file
        raise self.server.error(f"Retries exceeded for request: {url}")

    def wrap_request(self, default_url: str, **kwargs) -> HttpRequestWrapper:
        return HttpRequestWrapper(self, default_url, **kwargs)

    def close(self):
        self.client.close()

class HttpRequestWrapper:
    def __init__(
        self, client: HttpClient, default_url: str, **kwargs
    ) -> None:
        self._do_request = client.request
        self._last_response: Optional[HttpResponse] = None
        self.default_request_args: Dict[str, Any] = {
            "method": "GET",
            "url": default_url,
        }
        self.default_request_args.update(kwargs)
        self.request_args = copy.deepcopy(self.default_request_args)
        self.reset()

    async def send(self, **kwargs) -> HttpResponse:
        req_args = copy.deepcopy(self.request_args)
        req_args.update(kwargs)
        method = req_args.pop("method", self.default_request_args["method"])
        url = req_args.pop("url", self.default_request_args["url"])
        self._last_response = await self._do_request(method, url, **req_args)
        return self._last_response

    def set_method(self, method: str) -> None:
        self.request_args["method"] = method

    def set_url(self, url: str) -> None:
        self.request_args["url"] = url

    def set_body(
        self, body: Optional[Union[str, List[Any], Dict[str, Any]]]
    ) -> None:
        self.request_args["body"] = body

    def add_header(self, name: str, value: str) -> None:
        headers = self.request_args.get("headers", {})
        headers[name] = value
        self.request_args["headers"] = headers

    def set_headers(self, headers: Dict[str, str]) -> None:
        self.request_args["headers"] = headers

    def reset(self) -> None:
        self.request_args = copy.deepcopy(self.default_request_args)

    def last_response(self) -> Optional[HttpResponse]:
        return self._last_response

class HttpResponse:
    def __init__(self,
                 url: str,
                 final_url: str,
                 code: int,
                 result: bytes,
                 response_headers: HTTPHeaders,
                 error: Optional[BaseException]
                 ) -> None:
        self._url = url
        self._final_url = final_url
        self._code = code
        self._result: bytes = result
        self._encoding: str = "utf-8"
        self._response_headers: HTTPHeaders = response_headers
        self._etag: Optional[str] = response_headers.get("etag", None)
        self._error = error
        self._last_modified: Optional[str] = response_headers.get(
            "last-modified", None)

    def json(self) -> Union[List[Any], Dict[str, Any]]:
        return jsonw.loads(self._result)

    def is_cachable(self) -> bool:
        return self._last_modified is not None or self._etag is not None

    def has_error(self) -> bool:
        return self._error is not None

    def raise_for_status(self, message: Optional[str] = None) -> None:
        if self._error is not None:
            code = 500
            msg = f"HTTP Request Error: {self.url}"
            if isinstance(self._error, HTTPError):
                code = self._code
                if self._error.message is not None:
                    msg = self._error.message
            if message is not None:
                msg = message
            raise ServerError(msg, code) from self._error

    @property
    def encoding(self) -> str:
        return self._encoding

    @encoding.setter
    def encoding(self, new_enc: str) -> None:
        self._encoding = new_enc

    @property
    def text(self) -> str:
        return self._result.decode(encoding=self._encoding)

    @property
    def content(self) -> bytes:
        return self._result

    @property
    def url(self) -> str:
        return self._url

    @property
    def final_url(self) -> str:
        return self._final_url

    @property
    def status_code(self) -> int:
        return self._code

    @property
    def headers(self) -> HTTPHeaders:
        return self._response_headers

    @property
    def last_modified(self) -> Optional[str]:
        return self._last_modified

    @property
    def etag(self) -> Optional[str]:
        return self._etag

    @property
    def error(self) -> Optional[BaseException]:
        return self._error

class StreamingDownload:
    def __init__(
        self,
        server: Server,
        dest_path: Optional[StrOrPath],
        download_size: int,
        progress_callback: Optional[Callable[[int, int, int], None]]
    ) -> None:
        self.server = server
        self.event_loop = server.get_event_loop()
        self.need_content_length: bool = True
        self.need_content_disposition: bool = False
        self.request_ok: bool = False
        if dest_path is None:
            # If no destination is provided initialize to a procedurally
            # generated temp file.  We will attempt to extract the filename
            # from the Content-Disposition Header
            tmp_dir = tempfile.gettempdir()
            loop_time = int(self.event_loop.get_loop_time())
            tmp_fname = f"moonraker.download-{loop_time}.mrd"
            self.dest_file = pathlib.Path(tmp_dir).joinpath(tmp_fname)
            self.need_content_disposition = True
        elif isinstance(dest_path, str):
            self.dest_file = pathlib.Path(dest_path)
        else:
            self.dest_file = dest_path
        self.filename = self.dest_file.name
        self.file_hdl: Optional[BufferedWriter] = None
        self.total_recd: int = 0
        self.download_size: int = download_size
        self.pct_done: int = 0
        self.chunk_buffer: List[bytes] = []
        self.progress_callback = progress_callback
        self.busy_evt: asyncio.Event = asyncio.Event()
        self.busy_evt.set()

    def on_headers_recd(self, line: str) -> None:
        if not self.need_content_length and not self.need_content_disposition:
            return
        line = line.strip()
        rc_match = re.match(r"HTTP/\d.?\d? (\d+)", line)
        if rc_match is not None:
            self.request_ok = rc_match.group(1) == "200"
            return
        if not self.request_ok:
            return
        parts = line.split(":", 1)
        if len(parts) < 2:
            return
        hname = parts[0].strip().lower()
        hval = parts[1].strip()
        if hname == "content-length" and self.need_content_length:
            self.download_size = int(hval)
            self.need_content_length = False
            logging.debug(
                f"Content-Length header received: "
                f"size = {self.download_size}")
        elif (
            hname == "content-disposition" and
            self.need_content_disposition
        ):
            fnr = r"filename[^;\n=]*=(['\"])?(utf-8\'\')?([^\n;]*)(?(1)\1|)"
            matches: List[Tuple[str, str, str]] = re.findall(fnr, hval)
            is_utf8 = False
            for (_, encoding, fname) in matches:
                if encoding.startswith("utf-8"):
                    # Prefer the utf8 filename if included
                    self.filename = url_unescape(
                        fname, encoding="utf-8", plus=False)
                    is_utf8 = True
                    break
                self.filename = fname
            self.need_content_disposition = False
            # Use the filename extracted from the content-disposition header
            self.dest_file = self.dest_file.parent.joinpath(self.filename)
            logging.debug(
                "Content-Disposition header received: filename = "
                f"{self.filename}, utf8: {is_utf8}")

    def on_chunk_recd(self, chunk: bytes) -> None:
        if not chunk:
            return
        self.chunk_buffer.append(chunk)
        if not self.busy_evt.is_set():
            return
        self.busy_evt.clear()
        self.event_loop.register_callback(self._process_buffer)

    async def close(self):
        await self.busy_evt.wait()
        if self.file_hdl is not None:
            await self.event_loop.run_in_thread(self.file_hdl.close)

    async def _process_buffer(self):
        if self.file_hdl is None:
            self.file_hdl = await self.event_loop.run_in_thread(
                self.dest_file.open, "wb")
        while self.chunk_buffer:
            chunk = self.chunk_buffer.pop(0)
            await self.event_loop.run_in_thread(self.file_hdl.write, chunk)
            self.total_recd += len(chunk)
            if self.download_size > 0 and self.progress_callback is not None:
                pct = int(self.total_recd / self.download_size * 100 + .5)
                pct = min(100, pct)
                if pct != self.pct_done:
                    self.pct_done = pct
                    self.progress_callback(
                        pct, self.download_size, self.total_recd)
        self.busy_evt.set()


def load_component(config: ConfigHelper) -> HttpClient:
    return HttpClient(config)




# This file was autogenerated from pk-enum.h by pk-enum-converter.py
# on Fri Jan 21 16:01:47 2022 UTC
#
# License for original source:
#
# Copyright (C) 2007-2014 Richard Hughes <richard@hughsie.com>
#
# Licensed under the GNU Lesser General Public License Version 2.1
#
# This library is free software; you can redistribute it and/or
# modify it under the terms of the GNU Lesser General Public
# License as published by the Free Software Foundation; either
# version 2.1 of the License, or (at your option) any later version.
#
# This library is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# Lesser General Public License for more details.
#
# You should have received a copy of the GNU Lesser General Public
# License along with this library; if not, write to the Free Software
# Foundation, Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301 USA
#

from __future__ import annotations
import sys
from enum import Flag, auto

class PkFlag(Flag):
    @classmethod
    def from_pkstring(cls, pkstring: str):
        for name, member in cls.__members__.items():
            if member.pkstring == pkstring:
                return cls(member.value)
        # Return "unknown" flag
        return cls(1)

    @classmethod
    def from_index(cls, index: int):
        return cls(1 << index)

    @property
    def pkstring(self) -> str:
        if self.name is None:
            return " | ".join([f.pkstring for f in self])
        return self.name.lower().replace("_", "-")

    @property
    def desc(self) -> str:
        if self.name is None:
            return ", ".join([f.desc for f in self])
        description = self.name.lower().replace("_", " ")
        return description.capitalize()

    @property
    def index(self) -> int:
        return self.value.bit_length() - 1

    if sys.version_info < (3, 11):
        def __iter__(self):
            for i in range(self._value_.bit_length()):
                val = 1 << i
                if val & self._value_ == val:
                    yield self.__class__(val)


class Role(PkFlag):
    """
    What we were asked to do, this never changes for the lifetime of the
    transaction.
    Icons that have to represent the whole "aim" of the transaction will use
    these constants

    * Role.UNKNOWN:  Unknow request
    * Role.CANCEL:  Cancel transaction
    * Role.DEPENDS_ON:  Get package dependencies
    * Role.GET_DETAILS:  Get package details
    * Role.GET_FILES:
    * Role.GET_PACKAGES:  Get available packages
    * Role.GET_REPO_LIST:  Get repository list
    * Role.REQUIRED_BY:  Get packages required by given package
    * Role.GET_UPDATE_DETAIL:  Get update details
    * Role.GET_UPDATES:  Get available updates
    * Role.INSTALL_FILES:  Install package files
    * Role.INSTALL_PACKAGES:  Install packages
    * Role.INSTALL_SIGNATURE:  Install signature
    * Role.REFRESH_CACHE:  Refresh cache
    * Role.REMOVE_PACKAGES:  Remove packages
    * Role.REPO_ENABLE:  Enable repository
    * Role.REPO_SET_DATA:
    * Role.RESOLVE:  Resolve depdencies
    * Role.SEARCH_DETAILS:  Search for details
    * Role.SEARCH_FILE:  Search for file
    * Role.SEARCH_GROUP:  Search for group
    * Role.SEARCH_NAME:  Search for package name
    * Role.UPDATE_PACKAGES:  Update packages
    * Role.WHAT_PROVIDES:  Get what a package provides
    * Role.ACCEPT_EULA:  Accept an EULA
    * Role.DOWNLOAD_PACKAGES:  Download packages
    * Role.GET_DISTRO_UPGRADES:  Get available distribution upgrades
    * Role.GET_CATEGORIES:  Get available categories
    * Role.GET_OLD_TRANSACTIONS:  Get old transation information
    * Role.REPAIR_SYSTEM:  Repair system
    * Role.GET_DETAILS_LOCAL:  Get details on local package
    * Role.GET_FILES_LOCAL:  Get files provided by local package
    * Role.REPO_REMOVE:  Remove repository
    * Role.UPGRADE_SYSTEM:  Upgrade system
    """
    UNKNOWN = auto()
    CANCEL = auto()
    DEPENDS_ON = auto()
    GET_DETAILS = auto()
    GET_FILES = auto()
    GET_PACKAGES = auto()
    GET_REPO_LIST = auto()
    REQUIRED_BY = auto()
    GET_UPDATE_DETAIL = auto()
    GET_UPDATES = auto()
    INSTALL_FILES = auto()
    INSTALL_PACKAGES = auto()
    INSTALL_SIGNATURE = auto()
    REFRESH_CACHE = auto()
    REMOVE_PACKAGES = auto()
    REPO_ENABLE = auto()
    REPO_SET_DATA = auto()
    RESOLVE = auto()
    SEARCH_DETAILS = auto()
    SEARCH_FILE = auto()
    SEARCH_GROUP = auto()
    SEARCH_NAME = auto()
    UPDATE_PACKAGES = auto()
    WHAT_PROVIDES = auto()
    ACCEPT_EULA = auto()
    DOWNLOAD_PACKAGES = auto()
    GET_DISTRO_UPGRADES = auto()
    GET_CATEGORIES = auto()
    GET_OLD_TRANSACTIONS = auto()
    REPAIR_SYSTEM = auto()
    GET_DETAILS_LOCAL = auto()
    GET_FILES_LOCAL = auto()
    REPO_REMOVE = auto()
    UPGRADE_SYSTEM = auto()

class Status(PkFlag):
    """
    What status we are now; this can change for each transaction giving a
    status of what sort of thing is happening
    Icons that change to represent the current status of the transaction will
    use these constants
    If you add to these, make sure you add filenames in gpk-watch.c also

    A typical transaction will do:
    - schedule task
        WAIT
    - run task
        SETUP
    - wait for lock
        RUNNING

    This means that backends should run pk_backend_set_status (backend,
    PK_STATUS_ENUM_RUNNING)
    when they are ready to start running the transaction and after a lock has
    been got.

    * Status.UNKNOWN:  Unknown status
    * Status.WAIT:  Waiting
    * Status.SETUP:  Setting up
    * Status.RUNNING:  Running
    * Status.QUERY:
    * Status.INFO:
    * Status.REMOVE:  Removing
    * Status.REFRESH_CACHE:  Refreshing cache
    * Status.DOWNLOAD:  Downloading
    * Status.INSTALL:  Installing
    * Status.UPDATE:  Updating
    * Status.CLEANUP:  Cleaning up
    * Status.OBSOLETE:
    * Status.DEP_RESOLVE:  Resolving dependencies
    * Status.SIG_CHECK:  Checking signatures
    * Status.TEST_COMMIT:  Testing commit
    * Status.COMMIT:  Committing
    * Status.REQUEST:
    * Status.FINISHED:  Finished
    * Status.CANCEL:  Cancelling
    * Status.DOWNLOAD_REPOSITORY:  Downloading respository
    * Status.DOWNLOAD_PACKAGELIST:  Donwloading package list
    * Status.DOWNLOAD_FILELIST:  Downloading file list
    * Status.DOWNLOAD_CHANGELOG:  Downloading changelog information
    * Status.DOWNLOAD_GROUP:  Downloading group information
    * Status.DOWNLOAD_UPDATEINFO:  Downloading update information
    * Status.REPACKAGING:  Repackaging
    * Status.LOADING_CACHE:  Loading cache
    * Status.SCAN_APPLICATIONS:  Scanning for applications
    * Status.GENERATE_PACKAGE_LIST:  Generating package list
    * Status.WAITING_FOR_LOCK:  Waiting for lock
    * Status.WAITING_FOR_AUTH:  Waiting for authentication/authorization
    * Status.SCAN_PROCESS_LIST:  Scanning running processes
    * Status.CHECK_EXECUTABLE_FILES:  Checking executable files
    * Status.CHECK_LIBRARIES:  Checking libraries
    * Status.COPY_FILES:  Copying files
    * Status.RUN_HOOK:  Running package hook
    """
    UNKNOWN = auto()
    WAIT = auto()
    SETUP = auto()
    RUNNING = auto()
    QUERY = auto()
    INFO = auto()
    REMOVE = auto()
    REFRESH_CACHE = auto()
    DOWNLOAD = auto()
    INSTALL = auto()
    UPDATE = auto()
    CLEANUP = auto()
    OBSOLETE = auto()
    DEP_RESOLVE = auto()
    SIG_CHECK = auto()
    TEST_COMMIT = auto()
    COMMIT = auto()
    REQUEST = auto()
    FINISHED = auto()
    CANCEL = auto()
    DOWNLOAD_REPOSITORY = auto()
    DOWNLOAD_PACKAGELIST = auto()
    DOWNLOAD_FILELIST = auto()
    DOWNLOAD_CHANGELOG = auto()
    DOWNLOAD_GROUP = auto()
    DOWNLOAD_UPDATEINFO = auto()
    REPACKAGING = auto()
    LOADING_CACHE = auto()
    SCAN_APPLICATIONS = auto()
    GENERATE_PACKAGE_LIST = auto()
    WAITING_FOR_LOCK = auto()
    WAITING_FOR_AUTH = auto()
    SCAN_PROCESS_LIST = auto()
    CHECK_EXECUTABLE_FILES = auto()
    CHECK_LIBRARIES = auto()
    COPY_FILES = auto()
    RUN_HOOK = auto()

class Exit(PkFlag):
    """
    How the backend exited

    * Exit.UNKNOWN:  Unknown exit status
    * Exit.SUCCESS:  Backend exited successfully
    * Exit.FAILED:  Backend failed
    * Exit.CANCELLED:  Backend was cancelled
    * Exit.KEY_REQUIRED:  A repository encryption key needs installing
    * Exit.EULA_REQUIRED:  A EULA is required to be accepted
    * Exit.KILLED:  Backend was killed
    * Exit.MEDIA_CHANGE_REQUIRED:  Media change required
    * Exit.NEED_UNTRUSTED:
    * Exit.CANCELLED_PRIORITY:  Cancelled due to higher priority task
    * Exit.SKIP_TRANSACTION:
    * Exit.REPAIR_REQUIRED:  Package database requires repairing
    """
    UNKNOWN = auto()
    SUCCESS = auto()
    FAILED = auto()
    CANCELLED = auto()
    KEY_REQUIRED = auto()
    EULA_REQUIRED = auto()
    KILLED = auto()
    MEDIA_CHANGE_REQUIRED = auto()
    NEED_UNTRUSTED = auto()
    CANCELLED_PRIORITY = auto()
    SKIP_TRANSACTION = auto()
    REPAIR_REQUIRED = auto()

class Network(PkFlag):
    """
    Network type

    * Network.UNKNOWN:  Unknown network
    * Network.OFFLINE:  Offline (no network)
    * Network.ONLINE:  Online (network type unknown)
    * Network.WIRED:  Wired network
    * Network.WIFI:  WiFi network
    * Network.MOBILE:  Mobile network
    """
    UNKNOWN = auto()
    OFFLINE = auto()
    ONLINE = auto()
    WIRED = auto()
    WIFI = auto()
    MOBILE = auto()

class Filter(PkFlag):
    """
    The filter types

    * Filter.UNKNOWN:  Unknown filter
    * Filter.NONE:  No filter
    * Filter.INSTALLED:  Filter for installed packages
    * Filter.NOT_INSTALLED:  Filter for not installed packages
    * Filter.DEVELOPMENT:  Filter for development packages
    * Filter.NOT_DEVELOPMENT:  Filter for non-development packages
    * Filter.GUI:  Filter for GUI packages
    * Filter.NOT_GUI:  Filter for non-GUI packages
    * Filter.FREE:  Filter for free packages
    * Filter.NOT_FREE:  Filter for non-free packages
    * Filter.VISIBLE:  Filter for visible packages
    * Filter.NOT_VISIBLE:  Filter for invisible packages
    * Filter.SUPPORTED:  Filter for supported packages
    * Filter.NOT_SUPPORTED:  Filter for not supported packages
    * Filter.BASENAME:  Filter for packages that match basename
    * Filter.NOT_BASENAME:  Filter for packages that don't match basename
    * Filter.NEWEST:  Filter for newest packages
    * Filter.NOT_NEWEST:  Filter for not newest packages
    * Filter.ARCH:  Filter for packages that match architecture
    * Filter.NOT_ARCH:  Filter for packages that don't match architecture
    * Filter.SOURCE:  Filter for source packages
    * Filter.NOT_SOURCE:  Filter for non-source packages
    * Filter.COLLECTIONS:  Filter for collections
    * Filter.NOT_COLLECTIONS:  Filter for not collections
    * Filter.APPLICATION:  Filter for application packages
    * Filter.NOT_APPLICATION:  Filter for non-application packages
    * Filter.DOWNLOADED:  Filter for downloaded packages
    * Filter.NOT_DOWNLOADED:  Filter for not downloaded packages
    """
    @property
    def pkstring(self) -> str:
        pks = self.name
        if pks is None:
            return " | ".join([f.pkstring for f in self])
        if pks in ["DEVELOPMENT", "NOT_DEVELOPMENT"]:
            pks = pks[:-6]
        if pks[:4] == "NOT_":
            pks = "~" + pks[4:]
        return pks.lower().replace("_", "-")

    UNKNOWN = auto()
    NONE = auto()
    INSTALLED = auto()
    NOT_INSTALLED = auto()
    DEVELOPMENT = auto()
    NOT_DEVELOPMENT = auto()
    GUI = auto()
    NOT_GUI = auto()
    FREE = auto()
    NOT_FREE = auto()
    VISIBLE = auto()
    NOT_VISIBLE = auto()
    SUPPORTED = auto()
    NOT_SUPPORTED = auto()
    BASENAME = auto()
    NOT_BASENAME = auto()
    NEWEST = auto()
    NOT_NEWEST = auto()
    ARCH = auto()
    NOT_ARCH = auto()
    SOURCE = auto()
    NOT_SOURCE = auto()
    COLLECTIONS = auto()
    NOT_COLLECTIONS = auto()
    APPLICATION = auto()
    NOT_APPLICATION = auto()
    DOWNLOADED = auto()
    NOT_DOWNLOADED = auto()

class Restart(PkFlag):
    """
    What restart we need to after a transaction, ordered by severity

    * Restart.UNKNOWN:  Unknown restart state
    * Restart.NONE:  No restart required
    * Restart.APPLICATION:  Need to restart the application
    * Restart.SESSION:  Need to restart the session
    * Restart.SYSTEM:  Need to restart the system
    * Restart.SECURITY_SESSION:
    * Restart.SECURITY_SYSTEM:
    """
    UNKNOWN = auto()
    NONE = auto()
    APPLICATION = auto()
    SESSION = auto()
    SYSTEM = auto()
    SECURITY_SESSION = auto()
    SECURITY_SYSTEM = auto()

class Error(PkFlag):
    """
    The error type

    * Error.UNKNOWN:
    * Error.OOM:  Out of memory
    * Error.NO_NETWORK:  No network access available
    * Error.NOT_SUPPORTED:  Request not supported
    * Error.INTERNAL_ERROR:  Undefined internal error
    * Error.GPG_FAILURE:  GPG encryption failure
    * Error.PACKAGE_ID_INVALID:  Invalid package ID provided
    * Error.PACKAGE_NOT_INSTALLED:  Requested package not installed
    * Error.PACKAGE_NOT_FOUND:  Requested package not found
    * Error.PACKAGE_ALREADY_INSTALLED:  Requested package already installed
    * Error.PACKAGE_DOWNLOAD_FAILED:  Failed to download package
    * Error.GROUP_NOT_FOUND:  Requested group not gound
    * Error.GROUP_LIST_INVALID:  Invalid group list provided
    * Error.DEP_RESOLUTION_FAILED:  Failed to resolve dependencies
    * Error.FILTER_INVALID:  Invalid filter provides
    * Error.CREATE_THREAD_FAILED:  Failed to create thread
    * Error.TRANSACTION_ERROR:  Error occurred during transaction
    * Error.TRANSACTION_CANCELLED:  Transaction was cancelled
    * Error.NO_CACHE:  No cache available
    * Error.REPO_NOT_FOUND:  Requested repository not found
    * Error.CANNOT_REMOVE_SYSTEM_PACKAGE:  Not allowed to remove system package
    * Error.PROCESS_KILL:  Process killed
    * Error.FAILED_INITIALIZATION:
    * Error.FAILED_FINALISE:
    * Error.FAILED_CONFIG_PARSING:  Configuration is not valid
    * Error.CANNOT_CANCEL:
    * Error.CANNOT_GET_LOCK:  Cannot get lock
    * Error.NO_PACKAGES_TO_UPDATE:  No packages to update
    * Error.CANNOT_WRITE_REPO_CONFIG:  Cannot write repository configuration
    * Error.LOCAL_INSTALL_FAILED:
    * Error.BAD_GPG_SIGNATURE:  Bad GPG signature found
    * Error.MISSING_GPG_SIGNATURE:  Required GPG signature not found
    * Error.CANNOT_INSTALL_SOURCE_PACKAGE:  Cannot install source package
    * Error.REPO_CONFIGURATION_ERROR:
    * Error.NO_LICENSE_AGREEMENT:
    * Error.FILE_CONFLICTS:  File conflicts detected
    * Error.PACKAGE_CONFLICTS:  Package conflict
    * Error.REPO_NOT_AVAILABLE:  Repository not available
    * Error.INVALID_PACKAGE_FILE:
    * Error.PACKAGE_INSTALL_BLOCKED:  Package installation blocked
    * Error.PACKAGE_CORRUPT:  Package corruption occurred
    * Error.ALL_PACKAGES_ALREADY_INSTALLED:  All packages already installed
    * Error.FILE_NOT_FOUND:  Required file not found
    * Error.NO_MORE_MIRRORS_TO_TRY:  Out of repository mirrors to try
    * Error.NO_DISTRO_UPGRADE_DATA:  No distribution upgrade path found
    * Error.INCOMPATIBLE_ARCHITECTURE:  Incompatible architecture found
    * Error.NO_SPACE_ON_DEVICE:  Out of required disk space
    * Error.MEDIA_CHANGE_REQUIRED:  Need to change media
    * Error.NOT_AUTHORIZED:  Authorization failed
    * Error.UPDATE_NOT_FOUND:  Update not found
    * Error.CANNOT_INSTALL_REPO_UNSIGNED:
         Installation repository missing signature
    * Error.CANNOT_UPDATE_REPO_UNSIGNED:  Update repository missing signature
    * Error.CANNOT_GET_FILELIST:  Cannot get file list
    * Error.CANNOT_GET_REQUIRES:  Cannot get package requirements
    * Error.CANNOT_DISABLE_REPOSITORY:  Cannot disable reposoitory
    * Error.RESTRICTED_DOWNLOAD:
    * Error.PACKAGE_FAILED_TO_CONFIGURE:  Package failed to configure
    * Error.PACKAGE_FAILED_TO_BUILD:  Package failed to build
    * Error.PACKAGE_FAILED_TO_INSTALL:  Package failed to install
    * Error.PACKAGE_FAILED_TO_REMOVE:  Package failed to remove
    * Error.UPDATE_FAILED_DUE_TO_RUNNING_PROCESS:
    * Error.PACKAGE_DATABASE_CHANGED:
    * Error.PROVIDE_TYPE_NOT_SUPPORTED:
    * Error.INSTALL_ROOT_INVALID:  Installtion root not suitable
    * Error.CANNOT_FETCH_SOURCES:  Cannot fetch sources
    * Error.CANCELLED_PRIORITY:  Cancelled due to higher priority task
    * Error.UNFINISHED_TRANSACTION:  Transaction unfinished
    * Error.LOCK_REQUIRED:  Required lock not available
    * Error.REPO_ALREADY_SET:
    """
    @property
    def pkstring(self) -> str:
        if self == Error.UPDATE_FAILED_DUE_TO_RUNNING_PROCESS:
            return "failed-due-to-running-process"
        return super().pkstring

    UNKNOWN = auto()
    OUT_OF_MEMORY = auto()
    NO_NETWORK = auto()
    NOT_SUPPORTED = auto()
    INTERNAL_ERROR = auto()
    GPG_FAILURE = auto()
    PACKAGE_ID_INVALID = auto()
    PACKAGE_NOT_INSTALLED = auto()
    PACKAGE_NOT_FOUND = auto()
    PACKAGE_ALREADY_INSTALLED = auto()
    PACKAGE_DOWNLOAD_FAILED = auto()
    GROUP_NOT_FOUND = auto()
    GROUP_LIST_INVALID = auto()
    DEP_RESOLUTION_FAILED = auto()
    FILTER_INVALID = auto()
    CREATE_THREAD_FAILED = auto()
    TRANSACTION_ERROR = auto()
    TRANSACTION_CANCELLED = auto()
    NO_CACHE = auto()
    REPO_NOT_FOUND = auto()
    CANNOT_REMOVE_SYSTEM_PACKAGE = auto()
    PROCESS_KILL = auto()
    FAILED_INITIALIZATION = auto()
    FAILED_FINALISE = auto()
    FAILED_CONFIG_PARSING = auto()
    CANNOT_CANCEL = auto()
    CANNOT_GET_LOCK = auto()
    NO_PACKAGES_TO_UPDATE = auto()
    CANNOT_WRITE_REPO_CONFIG = auto()
    LOCAL_INSTALL_FAILED = auto()
    BAD_GPG_SIGNATURE = auto()
    MISSING_GPG_SIGNATURE = auto()
    CANNOT_INSTALL_SOURCE_PACKAGE = auto()
    REPO_CONFIGURATION_ERROR = auto()
    NO_LICENSE_AGREEMENT = auto()
    FILE_CONFLICTS = auto()
    PACKAGE_CONFLICTS = auto()
    REPO_NOT_AVAILABLE = auto()
    INVALID_PACKAGE_FILE = auto()
    PACKAGE_INSTALL_BLOCKED = auto()
    PACKAGE_CORRUPT = auto()
    ALL_PACKAGES_ALREADY_INSTALLED = auto()
    FILE_NOT_FOUND = auto()
    NO_MORE_MIRRORS_TO_TRY = auto()
    NO_DISTRO_UPGRADE_DATA = auto()
    INCOMPATIBLE_ARCHITECTURE = auto()
    NO_SPACE_ON_DEVICE = auto()
    MEDIA_CHANGE_REQUIRED = auto()
    NOT_AUTHORIZED = auto()
    UPDATE_NOT_FOUND = auto()
    CANNOT_INSTALL_REPO_UNSIGNED = auto()
    CANNOT_UPDATE_REPO_UNSIGNED = auto()
    CANNOT_GET_FILELIST = auto()
    CANNOT_GET_REQUIRES = auto()
    CANNOT_DISABLE_REPOSITORY = auto()
    RESTRICTED_DOWNLOAD = auto()
    PACKAGE_FAILED_TO_CONFIGURE = auto()
    PACKAGE_FAILED_TO_BUILD = auto()
    PACKAGE_FAILED_TO_INSTALL = auto()
    PACKAGE_FAILED_TO_REMOVE = auto()
    UPDATE_FAILED_DUE_TO_RUNNING_PROCESS = auto()
    PACKAGE_DATABASE_CHANGED = auto()
    PROVIDE_TYPE_NOT_SUPPORTED = auto()
    INSTALL_ROOT_INVALID = auto()
    CANNOT_FETCH_SOURCES = auto()
    CANCELLED_PRIORITY = auto()
    UNFINISHED_TRANSACTION = auto()
    LOCK_REQUIRED = auto()
    REPO_ALREADY_SET = auto()
    OOM = OUT_OF_MEMORY

class Group(PkFlag):
    """
    The group type

    * Group.UNKNOWN:  Unknown group
    * Group.ACCESSIBILITY:  Accessibility related packages
    * Group.ACCESSORIES:  Accessory packages
    * Group.ADMIN_TOOLS:  Administration tools packages
    * Group.COMMUNICATION:  Communication packages
    * Group.DESKTOP_GNOME:  GNOME packages
    * Group.DESKTOP_KDE:  KDE packages
    * Group.DESKTOP_OTHER:  Other desktop packages
    * Group.DESKTOP_XFCE:  XFCE packages
    * Group.EDUCATION:  Education packages
    * Group.FONTS:  Fonts
    * Group.GAMES:  Games
    * Group.GRAPHICS:  Graphics related packages
    * Group.INTERNET:  Internet related packages
    * Group.LEGACY:  Legacy packages
    * Group.LOCALIZATION:  Localization related packages
    * Group.MAPS:  Map related packages
    * Group.MULTIMEDIA:  Multimedia packages
    * Group.NETWORK:  Network related packages
    * Group.OFFICE:  Office packages
    * Group.OTHER:
    * Group.POWER_MANAGEMENT:  Power-management related packages
    * Group.PROGRAMMING:  Programming packages
    * Group.PUBLISHING:  Publishing related packages
    * Group.REPOS:
    * Group.SECURITY:  Security packages
    * Group.SERVERS:  Server related packages
    * Group.SYSTEM:  System packages
    * Group.VIRTUALIZATION:  Virtualization packages
    * Group.SCIENCE:  Science related packages
    * Group.DOCUMENTATION:  Documentation
    * Group.ELECTRONICS:  Electronics package
    * Group.COLLECTIONS:
    * Group.VENDOR:  Vendor defined group
    * Group.NEWEST:  Special group for recently updated packages
    """
    UNKNOWN = auto()
    ACCESSIBILITY = auto()
    ACCESSORIES = auto()
    ADMIN_TOOLS = auto()
    COMMUNICATION = auto()
    DESKTOP_GNOME = auto()
    DESKTOP_KDE = auto()
    DESKTOP_OTHER = auto()
    DESKTOP_XFCE = auto()
    EDUCATION = auto()
    FONTS = auto()
    GAMES = auto()
    GRAPHICS = auto()
    INTERNET = auto()
    LEGACY = auto()
    LOCALIZATION = auto()
    MAPS = auto()
    MULTIMEDIA = auto()
    NETWORK = auto()
    OFFICE = auto()
    OTHER = auto()
    POWER_MANAGEMENT = auto()
    PROGRAMMING = auto()
    PUBLISHING = auto()
    REPOS = auto()
    SECURITY = auto()
    SERVERS = auto()
    SYSTEM = auto()
    VIRTUALIZATION = auto()
    SCIENCE = auto()
    DOCUMENTATION = auto()
    ELECTRONICS = auto()
    COLLECTIONS = auto()
    VENDOR = auto()
    NEWEST = auto()

class UpdateState(PkFlag):
    """
    What state the update is in

    * UpdateState.UNKNOWN:  Update stability unknown
    * UpdateState.STABLE:  Update is a stable release
    * UpdateState.UNSTABLE:  Update is an unstable release
    * UpdateState.TESTING:  Update is a testing release
    """
    UNKNOWN = auto()
    STABLE = auto()
    UNSTABLE = auto()
    TESTING = auto()

class Info(PkFlag):
    """
    The enumerated types used in Package() - these have to refer to a specific
    package action, rather than a general state

    * Info.UNKNOWN:  Package status is unknown
    * Info.INSTALLED:  Package is installed
    * Info.AVAILABLE:  Package is available to be installed
    * Info.LOW:
    * Info.ENHANCEMENT:
    * Info.NORMAL:
    * Info.BUGFIX:
    * Info.IMPORTANT:
    * Info.SECURITY:
    * Info.BLOCKED:  Package is blocked
    * Info.DOWNLOADING:  Package is downloading
    * Info.UPDATING:  Package is updating
    * Info.INSTALLING:  Package is being installed
    * Info.REMOVING:  Package is being removed
    * Info.CLEANUP:  Package is running cleanup
    * Info.OBSOLETING:
    * Info.COLLECTION_INSTALLED:
    * Info.COLLECTION_AVAILABLE:
    * Info.FINISHED:
    * Info.REINSTALLING:  Package is being reinstalled
    * Info.DOWNGRADING:  Package is being downgraded
    * Info.PREPARING:  Package is preparing for installation/removal
    * Info.DECOMPRESSING:  Package is decompressing
    * Info.UNTRUSTED:
    * Info.TRUSTED:
    * Info.UNAVAILABLE:  Package is unavailable
    * Info.CRITICAL:  Update severity is critical; Since: 1.2.4
    """
    UNKNOWN = auto()
    INSTALLED = auto()
    AVAILABLE = auto()
    LOW = auto()
    ENHANCEMENT = auto()
    NORMAL = auto()
    BUGFIX = auto()
    IMPORTANT = auto()
    SECURITY = auto()
    BLOCKED = auto()
    DOWNLOADING = auto()
    UPDATING = auto()
    INSTALLING = auto()
    REMOVING = auto()
    CLEANUP = auto()
    OBSOLETING = auto()
    COLLECTION_INSTALLED = auto()
    COLLECTION_AVAILABLE = auto()
    FINISHED = auto()
    REINSTALLING = auto()
    DOWNGRADING = auto()
    PREPARING = auto()
    DECOMPRESSING = auto()
    UNTRUSTED = auto()
    TRUSTED = auto()
    UNAVAILABLE = auto()
    CRITICAL = auto()

class DistroUpgrade(PkFlag):
    """
    The distro upgrade status

    * DistroUpgrade.UNKNOWN:  Unknown disto upgrade state
    * DistroUpgrade.STABLE:  Upgraded to stable release
    * DistroUpgrade.UNSTABLE:  Upgraded to unstable release
    """
    UNKNOWN = auto()
    STABLE = auto()
    UNSTABLE = auto()

class SigType(PkFlag):
    """
    The signature type type

    * SigType.UNKNOWN:  Unkwown signature type
    * SigType.GPG:  GPG signature
    """
    UNKNOWN = auto()
    GPG = auto()

class MediaType(PkFlag):
    """
    The media type

    * MediaType.UNKNOWN:  Unknown media type
    * MediaType.CD:  Media is a CD
    * MediaType.DVD:  Media is a DVD
    * MediaType.DISC:  Media is a disc (not CD or DVD)
    """
    UNKNOWN = auto()
    CD = auto()
    DVD = auto()
    DISC = auto()

class Authorize(PkFlag):
    """
    The authorization result

    * Authorize.UNKNOWN:  Unknown authorization status
    * Authorize.YES:  Authorized
    * Authorize.NO:  Not authorized
    * Authorize.INTERACTIVE:  Interaction required for authorization
    """
    UNKNOWN = auto()
    YES = auto()
    NO = auto()
    INTERACTIVE = auto()

class UpgradeKind(PkFlag):
    """
    The type of distribution upgrade to perform

    * UpgradeKind.UNKNOWN:
    * UpgradeKind.MINIMAL:  Perform minimal upgrade
    * UpgradeKind.DEFAULT:  Perform default upgrade
    * UpgradeKind.COMPLETE:  Perform complete upgrade
    """
    UNKNOWN = auto()
    MINIMAL = auto()
    DEFAULT = auto()
    COMPLETE = auto()

class TransactionFlag(PkFlag):
    """
    The transaction flags that alter how the transaction is handled

    * TransactionFlag.NONE:  No transaction flag
    * TransactionFlag.ONLY_TRUSTED:  Only allow trusted packages
    * TransactionFlag.SIMULATE:  Simulate transaction
    * TransactionFlag.ONLY_DOWNLOAD:  Only download packages
    * TransactionFlag.ALLOW_REINSTALL:  Allow package reinstallation
    * TransactionFlag.JUST_REINSTALL:  Only allow package reinstallation
    * TransactionFlag.ALLOW_DOWNGRADE:  Allow packages to be downgraded
    """
    NONE = auto()
    ONLY_TRUSTED = auto()
    SIMULATE = auto()
    ONLY_DOWNLOAD = auto()
    ALLOW_REINSTALL = auto()
    JUST_REINSTALL = auto()
    ALLOW_DOWNGRADE = auto()
